from typing import Tuple, Callable
from functools import partial

from anndata import AnnData
import pandas as pd
import numpy as np
import scipy.sparse as sps
import scipy.spatial as scs

from decoupler._docs import docs
from decoupler._log import _log
from decoupler.pp.data import extract


@docs.dedent
def get_obsm(
    adata: AnnData,
    key: str
) -> AnnData:
    """
    Extracts values stored in ``.obsm`` as a new AnnData object.
    This allows to reuse ``scanpy`` functions to visualise enrichment scores.

    Parameters
    ----------
    %(adata)s
    %(key)s

    Returns
    -------
    New AnnData object with values of the provided key in ``.obsm`` in ``X``.
    """
    # Validate
    assert isinstance(adata, AnnData), 'adata must be anndata.AnnData'
    assert isinstance(key, str), 'key must be str'
    assert key in adata.obsm, f'key={key} must be in adata.obsm'
    # Generate new AnnData
    obs = adata.obs
    X_obsm = adata.obsm[key]
    if isinstance(X_obsm, pd.DataFrame):
        var = pd.DataFrame(index=X_obsm.columns)
        X = X_obsm.values
    else:
        var = pd.DataFrame(index=[f'{i}' for i in range(X_obsm.shape[1])])
        X = X_obsm
    uns = adata.uns
    obsm = adata.obsm
    odata = AnnData(
        X=X,
        obs=obs,
        var=var,
        uns=uns,
        obsm=obsm
    )
    return odata


@docs.dedent
def swap_layer(
    adata: AnnData,
    key: str,
    X_key: str | None = 'X',
    inplace: bool = False,
) -> None | AnnData:
    """
    Swaps an ``AnnData.X`` for a given layer key. Generates a new object by default.

    Parameters
    ----------
    %(adata)s
    key
        ``adata.AnnData.layers`` key to place in ``adata.AnnData.X``.
    X_key
        ``adata.AnnData.layers`` key where to move and store the original ``adata.AnnData.X``.
        If None, the original ``adata.AnnData.X`` is discarded.
    inplace
        If ``False``, return a copy. Otherwise, do operation inplace and return ``None``.

    Returns
    -------
    If ``inplace=False``, new ``AnnData`` object.
    """
    # Validate
    assert isinstance(adata, AnnData), 'adata must be anndata.AnnData'
    assert isinstance(key, str), 'key must be str'
    assert key in adata.layers, f'key={key} must be in adata.layers'
    assert isinstance(X_key, str) or X_key is None, 'X_key must be str or None'
    assert isinstance(inplace, bool), 'inplace must be bool'
    # Move layers
    cdata = None
    if inplace:
        if X_key is not None:
            adata.layers[X_key] = adata.X
        adata.X = adata.layers[key]
    else:
        cdata = adata.copy()
        if X_key is not None:
            cdata.layers[X_key] = cdata.X
        cdata.X = cdata.layers[key]
    return cdata


def _validate_X(
    X: np.ndarray | sps.csr_matrix,
    mode: str = 'sum',
    skip_checks: bool = False
) -> None:
    assert isinstance(skip_checks, bool), 'skip_checks must be bool'
    skip_checks = type(mode) is dict or callable(mode) or skip_checks
    if not skip_checks:
        if isinstance(X, sps.csr_matrix):
            any_neg = (X.data < 0).any()
        else:
            any_neg = (X < 0).any()
        assert not any_neg, 'Provided data contains negative values.\n \
        Check the parameters raw and layers to determine if you are selecting the correct matrix.\n \
        If negative values are to be expected, override this error by setting skip_checks=True.'
        if mode == 'sum':
            if isinstance(X, sps.csr_matrix):
                is_int = float(np.sum(X.data)).is_integer()
            else:
                is_int = float(np.sum(X)).is_integer()
            assert is_int, 'Provided data contains float (decimal) values.\n \
            Check the parameters raw and layers to determine if you are selecting the correct matrix.\n \
            If decimal values are to be expected, override this error by setting skip_checks=True'


def _validate_mode(
    mode: str | Callable = 'sum',
    verbose: bool = False,
) -> Callable:
    if mode == 'sum':
        func = partial(np.sum, axis=0)
    elif mode == 'mean':
        func = partial(np.mean, axis=0)
    elif mode == 'median':
        func = partial(np.median, axis=0)
    elif callable(mode):
        func = partial(np.apply_along_axis, mode, 0)
    m = f'Using function {func.func.__name__} to aggregate observations'
    _log(m, level='info', verbose=verbose)
    return func


def _sample_group_by(
    sample: str,
    group: str | None,
    obs: pd.DataFrame,
    verbose: bool = False,
):
    # Use one column if the same
    if sample == group:
        group = None
    if group is None:
        # Filter extra columns in obs
        cols = obs.groupby(sample, observed=True).nunique(dropna=False).eq(other=1).all(axis=0)
        cols = np.hstack([sample, cols[cols].index])
        obs = obs.loc[:, cols]
        # Get unique samples
        smples = obs[sample].unique()
        groups = None
        # Get number of obs
        n_rows = len(smples)
        gsize = 0
    else:
        # Check if extra grouping is needed
        if type(group) is list:
            joined_cols = '_'.join(group)
            obs[joined_cols] = obs[group[0]].str.cat(obs[group[1:]].astype('U'), sep='_')
            group = joined_cols
        # Filter extra columns in obs
        cols = obs.groupby([sample, group], observed=True).nunique(dropna=False).eq(other=1).all(axis=0)
        cols = np.hstack([sample, group, cols[cols].index])
        obs = obs.loc[:, cols]
        # Get unique samples and groups
        smples = np.unique(obs[sample].values)
        groups = np.unique(obs[group].values)
        # Get number of obs
        n_rows = len(smples) * len(groups)
        gsize = groups.size
    m = f'Generating {n_rows} profiles: {smples.size} samples x {gsize} groups'
    _log(m, level='info', verbose=verbose)
    return obs, group, smples, groups, n_rows


def _psbulk(
    n_rows: int,
    n_cols: int,
    X: np.ndarray | sps.csr_matrix,
    sample_col: str,
    groups_col: str | None,
    smples: np.ndarray,
    groups: np.ndarray,
    obs: pd.DataFrame,
    new_obs: pd.DataFrame,
    mode: Callable,
    verbose: bool = False,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    # Init empty variables
    psbulk = np.zeros((n_rows, n_cols))
    props = np.zeros((n_rows, n_cols))
    ncells = np.zeros(n_rows)
    counts = np.zeros(n_rows)
    # Iterate for each group and sample
    i = 0
    if groups_col is None:
        for smp in smples:
            # Write new meta-data
            tmp = obs[obs[sample_col] == smp].drop_duplicates().values
            new_obs.loc[smp, :] = tmp
            # Get cells from specific sample
            profile = X[(obs[sample_col] == smp).values]
            if isinstance(X, sps.csr_matrix):
                profile = profile.toarray()
            # Skip if few cells or not enough counts
            ncell = profile.shape[0]
            count = np.sum(profile)
            ncells[i] = ncell
            counts[i] = count
            m = f'sample={smp}\tcells={ncell}\tcounts={count}'
            _log(m, level='info', verbose=verbose)
            if ncell == 0 or count == 0:
                i +=1
                continue
            # Get prop of non zeros
            prop = (profile.astype(bool)).mean(axis=0)
            # Pseudo-bulk
            profile = mode(profile)
            # Append
            props[i] = prop
            psbulk[i] = profile
            i += 1
    else:
        for grp in groups:
            for smp in smples:
                # Write new meta-data
                index = smp + '_' + grp
                tmp = obs[(obs[sample_col] == smp) & (obs[groups_col] == grp)].drop_duplicates().values
                if tmp.shape[0] == 0:
                    tmp = obs[obs[sample_col] == smp].drop(columns=groups_col).drop_duplicates()
                    tmp[groups_col] = grp
                    tmp = tmp[obs.columns].values
                new_obs.loc[index, :] = tmp
                # Get cells from specific sample and group
                profile = X[((obs[sample_col] == smp) & (obs[groups_col] == grp)).values]
                if isinstance(X, sps.csr_matrix):
                    profile = profile.toarray()
                # Skip if few cells or not enough counts
                ncell = profile.shape[0]
                count = np.sum(profile)
                ncells[i] = ncell
                counts[i] = count
                m = f'group={grp}\tsample={smp}\tcells={ncell}\tcounts={count}'
                _log(m, level='info', verbose=verbose)
                if ncell == 0 or count == 0:
                    i +=1
                    continue
                # Get prop of non zeros
                prop = (profile.astype(bool)).mean(axis=0)
                # Pseudo-bulk
                profile = mode(profile)
                # Append
                props[i] = prop
                psbulk[i] = profile
                i += 1
    return psbulk, ncells, counts, props


@docs.dedent
def pseudobulk(
    adata: AnnData,
    sample_col: str,
    groups_col: str | None,
    layer: str | None = None,
    raw: bool = False,
    empty: bool = True,
    mode: str | Callable | dict = 'sum',
    skip_checks: bool = False,
    verbose: bool = False,
) -> AnnData:
    """
    Summarizes omic profiles across cells, grouped by sample and optionally by group categories.

    By default this function expects raw integer counts as input and sums them per sample and group (``mode='sum'``),
    but other modes are available.

    This function produces some quality control metrics to assess if is necessary to filter some samples or features.
    The number of cells that belong to each sample is stored in ``adata.obs['psbulk_n_cells']``,
    the total sum of counts per sample in ``.obs['psbulk_counts']``,
    and the proportion of cells that have a non-zero value for a given feature in ``.layers['psbulk_props']``.

    Parameters
    ----------
    %(adata)s
    sample_col
        Column of ``adata.obs`` where to extract the samples names.
    groups_col
        Column of ``adata.obs`` where to extract the groups names. Can be set to ``None`` to ignore groups.
    %(layer)s
    %(raw)s
    %(empty)s
    mode
        How to perform the pseudobulk. Available options are ``sum``, ``mean`` or ``median``. It also accepts callback
        functions, like lambda, to perform custom aggregations. Additionally, it is also possible to provide a dictionary of
        different callback functions, each one stored in a different resulting `.layer`.
        In this case, the result of the first callback function of the dictionary is stored in ``.X`` by default. To switch
        between layers check ``decoupler.swap_layer``.
    skip_checks
        Whether to skip input checks. Set to ``True`` when working with positive and negative data, or when counts are not
        integers and ``mode='sum'``.
    %(verbose)s

    Returns
    -------
    New AnnData object containing summarized pseudobulk profiles by sample and optionally by group.
    """
    # Validate
    assert isinstance(adata, AnnData), 'adata must be an AnnData instance'
    assert isinstance(sample_col, str), 'sample_col must be a str'
    assert isinstance(groups_col, str) or groups_col is None, 'sample_col must be str or None'
    assert isinstance(mode, (str, dict)) or callable(mode), 'mode must be str, dict or callable'
    # Extract data
    X, _, var = extract(adata, layer=layer, raw=raw, empty=empty, verbose=verbose)
    obs = adata.obs.copy()
    var = adata.var.loc[var]
    # Validate X
    _validate_X(X=X, mode=mode, skip_checks=skip_checks)
    # Format inputs
    obs, groups_col, smples, groups, n_rows = _sample_group_by(
        sample=sample_col,
        group=groups_col,
        obs=obs,
        verbose=verbose,
    )
    n_cols = var.index.size
    new_obs = pd.DataFrame(columns=obs.columns)
    if type(mode) is dict:
        psbulks = []
        for l_name in mode:
            func = mode[l_name]
            func = _validate_mode(mode=func, verbose=verbose)
            psbulk, ncells, counts, props = _psbulk(
                n_rows=n_rows,
                n_cols=n_cols,
                X=X,
                sample_col=sample_col,
                groups_col=groups_col,
                smples=smples,
                groups=groups,
                obs=obs,
                new_obs=new_obs,
                mode=func,
                verbose=verbose,
            )
            psbulks.append(psbulk)
        layers = {k: v for k, v in zip(mode.keys(), psbulks)}
        layers['psbulk_props'] = props
    else:
        # Compute psbulk
        func = _validate_mode(mode=mode, verbose=verbose)
        psbulk, ncells, counts, props = _psbulk(
            n_rows=n_rows,
            n_cols=n_cols,
            X=X,
            sample_col=sample_col,
            groups_col=groups_col,
            smples=smples,
            groups=groups,
            obs=obs,
            new_obs=new_obs,
            mode=func,
            verbose=verbose,
        )
        layers = {'psbulk_props': props}
    # Add QC metrics
    new_obs['psbulk_cells'] = ncells
    new_obs['psbulk_counts'] = counts
    # Make cats
    for col in new_obs.columns:
        if not pd.api.types.is_numeric_dtype(new_obs[col]):
            new_obs[col] = new_obs[col].astype('category')
    # Create new AnnData
    psbulk = AnnData(X=psbulk, obs=new_obs, var=var, layers=layers)
    # Place first element of mode dict as X
    if type(mode) is dict:
        swap_layer(psbulk, key=list(mode.keys())[0], X_key=None, inplace=True)
    return psbulk


def filter_samples(
    adata: AnnData,
    min_cells: int | float = 10,
    min_counts: int | float = 1000,
    inplace: bool = True,
) -> None | np.ndarray:
    """
    Remove pseudobulked samples with insufficient number of cells and total counts.

    Parameters
    ----------
    %(adata)s
    %(min_cells)s
    %(min_counts)s
    %(inplace)s

    Returns
    -------
    If ``inplace=False``, array of samples to be kept.
    """
    assert isinstance(adata, AnnData), 'adata must be AnnData'
    cols = {'psbulk_cells', 'psbulk_counts'}
    assert cols.issubset(adata.obs.columns), \
    'psbulk_cells and psbulk_counts must be in obs, ' \
    'run again decoupler.pp.pseudobulk'
    assert isinstance(min_cells, (int, float)) and min_cells > 0., \
    'min_cells must be numerical and bigger than 0'
    assert isinstance(min_counts, (int, float)), \
    'min_counts must be numerical'
    msk_cells = adata.obs['psbulk_cells'] >= min_cells
    msk_counts = adata.obs['psbulk_counts'] >= min_counts
    msk = msk_cells & msk_counts
    obs_names = adata.obs_names[msk].to_list()
    if inplace:
        adata._inplace_subset_obs(obs_names)
    else:
        return np.array(obs_names, dtype='U')


def _min_sample_size(
    group: str | None,
    obs: pd.DataFrame,
    large_n: int,
    min_prop: float,
) -> float:
    assert isinstance(group, str) or group is None, 'group must be str or None'
    if group is None:
        min_sample_size = obs.shape[0]
    else:
        min_sample_size = obs[group].value_counts().min()
    if min_sample_size > large_n:
        min_sample_size = large_n + (min_sample_size - large_n) * min_prop
    return min_sample_size


def _cpm_cutoff(
    lib_size: np.ndarray,
    min_count: float,
) -> float:
    median_lib_size = np.median(lib_size)
    cpm_cutoff = min_count / median_lib_size * 1e6
    return cpm_cutoff


def _cpm(
    X: np.ndarray,
    lib_size: float | np.ndarray,
) -> np.ndarray:
    if isinstance(lib_size, (int, float)):
        cpm = X / lib_size * 1e6
    else:
        cpm = X / lib_size.reshape(-1, 1) * 1e6
    return cpm

def _ssize_tcount(
    X: np.ndarray,
    lib_size: float | None = None,
    min_count: int = 10,
) -> Tuple[np.ndarray, np.ndarray]:
    if isinstance(X, sps.csr_matrix):
        X = X.toarray()
    # Compute lib_size if needed
    if lib_size is None:
        lib_size = np.sum(X, axis=1)
    # CPM cutoff
    cpm_cutoff = _cpm_cutoff(lib_size=lib_size, min_count=min_count)
    # CPM mask
    cpm = _cpm(X=X, lib_size=lib_size)
    sample_size = np.round(np.sum(cpm >= cpm_cutoff.reshape(-1, 1), axis=0))
    total_count = np.sum(X, axis=0)
    return sample_size, total_count


@docs.dedent
def filter_by_expr(
    adata: AnnData,
    group: str | None = None,
    lib_size: float | None = None,
    min_count: int = 10,
    min_total_count: int = 15,
    large_n: int = 10,
    min_prop: float = 0.7,
    inplace: bool = True,
) -> None | np.ndarray:
    """
    Determine which genes have sufficiently large counts to be retained in a statistical analysis.

    Adapted from the function ``filterByExpr`` of edgeR (https://rdrr.io/bioc/edgeR/man/filterByExpr.html).

    Parameters
    ----------
    %(adata)s
    group
        Name of the ``adata.obs`` column to group by. If None, it assumes that all samples belong to one group.
    lib_size
        Library size. If None, default to the sum of reads per sample.
    min_count
        Minimum count requiered per gene for at least some samples.
    min_total_count
        Minimum total count required per gene across all samples.
    large_n
        Number of samples per group that is considered to be "large".
    min_prop
        Minimum proportion of samples in the smallest group that express the gene.
    %(inplace)s

    Returns
    -------
    If ``inplace=False``, array of genes to be kept.
    """
    # Validate
    assert isinstance(adata, AnnData), 'adata must be AnnData'
    assert isinstance(lib_size, (int, float)) or lib_size is None, \
    'lib_size must be numeric or None'
    assert isinstance(min_count, (int, float)) and min_count >= 0, \
    'min_count must be numeric and > 0'
    assert isinstance(min_total_count, (int, float)) and min_total_count >= 0, \
    'min_total_count must be numeric and > 0'
    assert isinstance(large_n, (int, float)) and large_n >= 0, \
    'large_n must be numeric and > 0'
    assert isinstance(min_prop, (int, float)) and 1 >= min_prop >= 0, \
    'min_prop must be numeric and between 0 and 1'
    # Extract inputs
    X, _, var_names = extract(adata, empty=False)
    obs = adata.obs
    # Minimum sample size cutoff
    min_sample_size = _min_sample_size(
        obs=obs,
        group=group,
        large_n=large_n,
        min_prop=min_prop,
    )
    # Compute sample size and total count
    sample_size, total_count = _ssize_tcount(
        X=X,
        lib_size=lib_size,
        min_count=min_count,
    )
    # Sample size msk
    keep_cpm = sample_size >= (min_sample_size - 1e-14)
    # Total counts msk
    keep_total_count = total_count >= (min_total_count - 1e-14)
    # Merge msks
    msk = keep_cpm & keep_total_count
    genes = var_names[msk]
    if inplace:
        adata._inplace_subset_var(genes)
    else:
        return genes


@docs.dedent
def filter_by_prop(
    adata: AnnData,
    min_prop: float = 0.2,
    min_smpls: int = 2,
    inplace: bool = True,
) -> None | np.ndarray:
    """
    Determine which genes are expressed in a sufficient proportion of cells across samples.

    This function selects genes that are sufficiently expressed across cells in each sample and that this condition is
    met across a minimum number of samples.

    Parameters
    ----------
    %(adata)s
    min_prop
        Minimum proportion of cells that express a gene in a sample.
    min_smpls
        Minimum number of samples with bigger or equal proportion of cells with expression than ``min_prop``.
    %(inplace)s

    Returns
    -------
    If ``inplace=False``, array of genes to be kept.
    """
    # Validate
    assert isinstance(adata, AnnData), 'adata must be AnnData'
    assert 'psbulk_props' in adata.layers.keys(), \
    'psbulk_props must be in adata.layers, use this function afer running decoupler.pp.pseudobulk'
    assert isinstance(min_prop, (int, float)) and 1 >= min_prop >= 0, \
    'min_prop must be numeric and between 0 and 1'
    assert isinstance(min_smpls, (int, float)) and adata.obs_names.size >= min_smpls >= 0, \
    f'min_smpls must be numeric and {adata.obs_names.size} > 0'
    # Extract props
    props = adata.layers['psbulk_props']
    if isinstance(props, pd.DataFrame):
        props = props.values
    # Compute n_smpl
    nsmpls = np.sum(props >= min_prop, axis=0)
    # Set features to 0
    msk = nsmpls >= min_smpls
    genes = adata.var_names[msk]
    if inplace:
        adata._inplace_subset_var(genes)
    else:
        return np.array(genes, dtype='U')


@docs.dedent
def knn(
    adata: AnnData,
    key: str = 'spatial',
    bw: float = 100,
    max_nn: int = 100,
    cutoff: float = 0.1,
) -> None:
    """
    Adds K-Nearest Neighbors similarities based on spatial distances.

    Parameters
    ----------
    %(adata)s
    %(key)s
    bw
        Bandwith of kernel.
    max_nn
        Maximum number of nearest neighbors to consider.
    cutoff
        Values below this number are set to zero.
    """
    # Validate
    assert isinstance(adata, AnnData), 'adata must be anndata.AnnData'
    assert key in adata.obsm, \
    f'adata.obsm must contain the spatial coordinates in adata.obsm["{key}"]'
    assert isinstance(bw, (int, float)) and bw > 0, 'bw must be numeric and > 0'
    assert isinstance(max_nn, int) and max_nn > 0, 'max_nn must be int and > 0'
    assert isinstance(cutoff, (int, float)) and cutoff > 0, \
    'cutoff must be numeric and > 0'
    # Find NN and their eucl dists
    coords = adata.obsm[key]
    nobs = coords.shape[0]
    tree = scs.KDTree(coords)
    max_nn = np.min([max_nn + 1, adata.n_obs])
    dist, idx = tree.query(coords, k=max_nn, workers=4)
    # Gaussian
    dist = np.exp(-(dist ** 2.0) / (2.0 * bw ** 2.0))
    if cutoff is not None:
        dist = dist * (dist > cutoff)
    # L1 norm
    dist = dist / np.sum(np.abs(dist), axis=1).reshape(-1, 1)
    # Build sparse matrix
    krnl = sps.csr_matrix(
        (dist.ravel(), (np.repeat(np.arange(nobs), max_nn), idx.ravel())),
        shape=(nobs, nobs)
    )
    krnl.eliminate_zeros()
    # Store
    adata.obsp[f'{key}_connectivities'] = krnl


@docs.dedent
def bin_order(
    adata: AnnData,
    order: str,
    names: str | list = None,
    label: str | None = None,
    nbins: int = 100,
) -> pd.DataFrame:
    """
    Bins features along a continuous, ordered process such as pseudotime.

    Parameters
    ----------
    %(adata)s
    %(order)s
    names
        Name of features to bin.
    %(label)s
    %(nbins)s

    Returns
    -------
    DataFrame with sources binned alng a continous ordered proess.
    """
    # Validate
    assert isinstance(adata, AnnData), 'adata must be anndata.AnnData'
    assert isinstance(order, str) and order in adata.obs.columns, \
    'order must be str and in adata.obs.columns'
    assert isinstance(names, (str, list)) or names is None, \
    'names must be str, list or None'
    assert (isinstance(label, str) and label in adata.obs.columns) or label is None, \
    'label must be str and in adata.obs.columns, or None'
    assert nbins > 1 and isinstance(nbins, int), 'nbins should be higher than 1 and be an integer'
    # Get vars and ordinal variable
    if names is None:
        names = adata.var_names
    elif isinstance(names, str):
        names = [names]
    X = adata[:, names].X
    if sps.issparse(X):
        X = X.toarray()
    y = adata.obs[order].values
    # Normalize to 0 and 1
    y = np.abs(y) / np.abs(y).max()
    # Make windows
    bin_edges = np.linspace(0, 1, nbins + 1)
    bin_midpoints = (bin_edges[:-1] + bin_edges[1:]) / 2
    # Prepare label colors
    cols = ['name', 'midpoint', 'value']
    if label is not None:
        adata.obs[label] = pd.Categorical(adata.obs[label])
        if f'{label}_colors' not in adata.uns.keys():
            from matplotlib.colors import to_hex
            import matplotlib.pyplot as plt
            cmap = plt.get_cmap('tab10')
            adata.uns[f'{label}_colors'] = [to_hex(cmap(i)) for i in adata.obs[label].sort_values().cat.codes.unique()]
        cols += ['label', 'color']
    dfs = []
    for i, name in enumerate(names):
        # Assign to windows based on order
        df = pd.DataFrame()
        df['value'] = X[:, i].ravel()
        df['name'] = name
        df['order'] = y
        df['window'] = pd.cut(df['order'], bins=bin_edges, labels=False, include_lowest=True, right=True)
        df['midpoint'] = df['window'].map(lambda x: bin_midpoints[int(x)])
        if label is not None:
            df['label'] = adata.obs[label].values
            df['color'] = [adata.uns[f'{label}_colors'][i] for i in adata.obs[label].cat.codes]
        df = df.sort_values('order')
        dfs.append(df)
    df = pd.concat(dfs)
    df = df[cols]
    df = df.rename(columns={'midpoint': 'order'}).reset_index(drop=True)
    df['order'] = df['order'] / df['order'].max()
    return df



import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import pytest

import decoupler as dc


@pytest.mark.parametrize(
    'source,label,vmin,vmax',
    [
        ['T1', None, 0, 10],
        ['T2', 'group', -3, 10],
        ['T3', 'group', -20, 15],
        ['T4', 'group', -1, 20],
        ['T5', 'group', -2, 14],
    ]
)
def test_order_targets(
    tdata,
    net,
    source,
    label,
    vmin,
    vmax,
):
    dc.mt.ulm(tdata, net, tmin=0)
    fig = dc.pl.order_targets(
        adata=tdata,
        net=net,
        order='pstime',
        source=source,
        label=label,
        vmin=vmin,
        vmax=vmax,
        return_fig=True,
    )
    assert isinstance(fig, Figure)
    plt.close(fig)



import logging

import numpy as np
import statsmodels.api as sm
import pytest

import decoupler as dc


def test_func_mlm(
    mat,
    adjmat,
):
    X, obs, var = mat
    dc_es, dc_pv = dc.mt._mlm._func_mlm(mat=X, adj=adjmat)
    st_es, st_pv = np.zeros(dc_es.shape), np.zeros(dc_pv.shape)
    for i in range(st_es.shape[0]):
        y = X[i, :]
        x = sm.add_constant(adjmat)
        model = sm.OLS(y, x)
        res = model.fit()
        st_es[i, :] = res.tvalues[1:]
        st_pv[i, :] = res.pvalues[1:]
    assert np.allclose(dc_es, st_es)
    assert np.allclose(dc_pv, st_pv)



from typing import Tuple

import numpy as np
import scipy.stats as sts

from decoupler.bm._pp import _validate_bool


def qrank(
    y_true: np.ndarray,
    y_score: np.ndarray,
) -> Tuple[float, float]:
    """
    1 - quantile normalized rank
    """
    _validate_bool(y_true=y_true, y_score=y_score)
    y_rank = sts.rankdata(y_score, axis=1, nan_policy='omit', method='average')
    y_rank = y_rank / np.sum(~np.isnan(y_rank), axis=1).reshape(-1, 1)
    msk = y_true.astype(np.bool_)
    score = y_rank[msk]
    rest = y_rank[~msk]
    _, pval = sts.ranksums(score, rest, alternative='greater')
    score = np.nanmean(score)
    return score, -np.log10(pval)

qrank.scores = ['1-qrank', '-log10(pval)']



import numpy as np
import pandas as pd

from decoupler._docs import docs
from decoupler._log import _log
from decoupler.op._resource import resource


@docs.dedent
def progeny(
    organism: str = 'human',
    top: int | float = np.inf,
    thr_padj: float = 0.05,
    license: str = 'academic',
    verbose: bool = False,
) -> pd.DataFrame:
    """
    Pathway RespOnsive GENes for activity inference (PROGENy) :cite:p:`progeny`.

    Wrapper to access PROGENy model gene weights. Each pathway is defined with
    a collection of target genes, each interaction has an associated p-value
    and weight. The top significant interactions per pathway are returned.

    Here is a brief description of each pathway:

    - **Androgen**: involved in the growth and development of the male reproductive organs
    - **EGFR**: regulates growth, survival, migration, apoptosis, proliferation, and differentiation in mammalian cells
    - **Estrogen**: promotes the growth and development of the female reproductive organs
    - **Hypoxia**: promotes angiogenesis and metabolic reprogramming when O2 levels are low
    - **JAK-STAT**: involved in immunity, cell division, cell death, and tumor formation
    - **MAPK**: integrates external signals and promotes cell growth and proliferation
    - **NFkB**: regulates immune response, cytokine production and cell survival
    - **p53**: regulates cell cycle, apoptosis, DNA repair and tumor suppression
    - **PI3K**: promotes growth and proliferation
    - **TGFb**: involved in development, homeostasis, and repair of most tissues
    - **TNFa**: mediates haematopoiesis, immune surveillance, tumour regression and protection from infection
    - **Trail**: induces apoptosis
    - **VEGF**: mediates angiogenesis, vascular permeability, and cell migration
    - **WNT**: regulates organ morphogenesis during development and tissue repair

    Parameters
    ----------
    %(organism)s
    top
        Number of genes per pathway to return. By default all of them.
    thr_padj
        Significance threshold to trim interactions.
    %(license)s
    %(verbose)s

    Returns
    -------
    Dataframe in long format containing target genes for each pathway with their associated weights and p-values.
    """
    # Validate
    assert isinstance(top, (int, float)) and top > 0, \
    'top must be numeric and > 0'
    assert isinstance(thr_padj, (int, float)) and 0. <= thr_padj <= 1., \
    'thr_padj must be numeric and between 0 and 1'
    # Download
    p = resource(name='PROGENy', organism=organism, license=license, verbose=verbose)
    p = (
        p
        .sort_values('p_value')
        .groupby('pathway')
        .head(top)
        .sort_values(['pathway', 'p_value'])
        .reset_index(drop=True)
    )
    p = p.rename(columns={'pathway': 'source', 'genesymbol': 'target', 'p_value': 'padj'})
    p = p[p['padj'] < thr_padj]
    p = p[['source', 'target', 'weight', 'padj']]
    m = f'progeny - filtered interactions for padj < {thr_padj}'
    _log(m, level='info', verbose=verbose)
    p = p.drop_duplicates(['source', 'target']).reset_index(drop=True)
    return p



import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import pytest

import decoupler as dc


@pytest.mark.parametrize(
    'names,label,mode',
    [
        [['G01', 'G02', 'G07', 'G08', 'G12'], None, 'line'],
        [['G01', 'G02', 'G07', 'G08'], None, 'mat'],
        [None, 'group', 'line'],
        [None, 'group', 'mat'],
    ]
)
def test_order(
    tdata,
    names,
    label,
    mode,
):
    df = dc.pp.bin_order(adata=tdata, names=['G12', 'G01', 'G07', 'G04'], order='pstime', label=label)
    fig = dc.pl.order(df=df, mode=mode, return_fig=True)
    assert isinstance(fig, Figure)
    plt.close(fig)



import numpy as np
import pytest

import decoupler as dc


def test_consensus(
    adata,
    net,
):
    dc.mt.decouple(data=adata, net=net, methods=['zscore', 'ulm'], cons=False, tmin=0)
    dc.mt.consensus(adata)
    assert 'score_consensus' in adata.obsm
    res = dc.mt.decouple(data=adata.to_df(), net=net, methods=['zscore', 'ulm'], cons=False, tmin=0)
    es, pv = dc.mt.consensus(res)
    assert np.isfinite(es.values).all()
    assert ((0 <= pv.values) & (pv.values <= 1)).all() 



from anndata import AnnData

from decoupler._docs import docs
from decoupler._log import _log
from decoupler._download import _download


@docs.dedent
def hsctgfb(
    verbose: bool = False,
) -> AnnData:
    """
    Downloads RNA-seq bulk data consisting of 6 samples of hepatic stellate cells
    (HSC) where three of them were activated by the cytokine
    Transforming growth factor (TGF-β) :cite:`hsc_tgfb`.

    Parameters
    ----------
    %(verbose)s

    Returns
    -------
    AnnData object.
    """
    # Download
    url = (
        'https://www.ncbi.nlm.nih.gov/geo/download/?acc=GSE151251&format=file&'
        'file=GSE151251%5FHSCs%5FCtrl%2Evs%2EHSCs%5FTGFb%2Ecounts%2Etsv%2Egz'
    )
    adata = _download(url, compression='gzip', sep='\t', verbose=verbose)
    # Transform to AnnData
    adata = adata.drop_duplicates('GeneName').set_index('GeneName').iloc[:, 5:].T
    adata.columns.name = None
    adata = AnnData(adata)
    adata.X = adata.X.astype(float)
    # Format obs
    adata.obs['condition'] = ['control' if '-Ctrl' in sample_id else 'treatment' for sample_id in adata.obs.index]
    adata.obs['sample_id'] = [sample_id.split('_')[0] for sample_id in adata.obs.index]
    adata.obs['condition'] = adata.obs['condition'].astype('category')
    adata.obs['sample_id'] = adata.obs['sample_id'].astype('category')
    m = f'generated AnnData with shape={adata.shape}'
    _log(m, level='info', verbose=verbose)
    return adata


@docs.dedent
def knocktf(
    thr_fc: int | float | None = -1,
    verbose: bool = False,
) -> AnnData:
    """
    Downloads gene contrast statistics from KnockTF :cite:`knocktf`,
    a large collection of transcription factor (TF) RNA-seq
    perturbation experiments.

    The values in ``adata.X`` represent the log2FCs of genes between
    perturbed and unperturbed samples.

    It also downloads all metadata associated to each perturbation
    experiment, such as which TF was perturbed, or in which tissue.

    Parameters
    ----------
    %(verbose)s

    Returns
    -------
    AnnData object.
    """
    assert isinstance(thr_fc, (int, float)) or thr_fc is None, \
    'thr_fc must be numeric or None'
    # Download
    url = (
        'https://zenodo.org/record/7035528/'
        'files/knockTF_expr.csv?download=1'
    )
    adata = _download(url, sep=',', index_col=0, verbose=verbose)
    url = (
        'https://zenodo.org/record/7035528/'
        'files/knockTF_meta.csv?download=1'
    )
    obs = _download(url, sep=',', index_col=0, verbose=verbose)
    obs = obs.rename(columns={'TF': 'source'}).assign(type_p=-1)
    # Make anndata
    adata = AnnData(X=adata, obs=obs)
    # Filter by thr_fc
    if thr_fc is not None:
        msk = adata.obs['logFC'] < thr_fc
        prc_keep = (msk.sum()/msk.size) * 100
        m = f'filtering AnnData for thr_fc={thr_fc}, will keep {prc_keep:.2f}% of observations'
        _log(m, level='info', verbose=verbose)
        adata = adata[msk, :].copy()
    m = f'generated AnnData with shape={adata.shape}'
    _log(m, level='info', verbose=verbose)
    return adata



import pandas as pd
import numpy as np
from matplotlib.figure import Figure
import adjustText as at

from decoupler._docs import docs
from decoupler._Plotter import Plotter
from decoupler.pp.net import _validate_net


@docs.dedent
def source_targets(
    data: pd.DataFrame,
    net: pd.DataFrame,
    x: str,
    y: str,
    name: str,
    top: int = 5,
    thr_x: float = 0.,
    thr_y: float = 0.,
    max_x: float | None = None,
    max_y: float | None = None,
    color_pos: str = '#D62728',
    color_neg: str = '#1F77B4',
    **kwargs,
) -> None | Figure:
    """
    Plots target features of a given source as a scatter plot.

    Parameters
    ----------
    %(data_plot)s
    %(net)s
    x
        Name of the column containing values to place on the x-axis.
    y
        Name of the column containing values to place on the y-axis.
    name
        Name of the source to plot.
    top
        Number of top features based on the product of x and y to label.
    thr_x
        Value were to place a baseline for the x-axis.
    thr_y
        Value were to place a baseline for the y-axis.
    max_x
        Maximum value to plot on x-axis.
    max_y
        Maximum value to plot on y-axis.
    color_pos
        Color to plot positively associated features.
    color_neg
        Color to plot negatively associated features.
    %(plot)s
    """
    # Validate inputs
    m = f'data must be a pd.DataFrame containing the columns {x} and {y}'
    assert isinstance(data, pd.DataFrame), m
    assert {x, y}.issubset(data.columns.union(net.columns)), m
    assert not pd.api.types.is_numeric_dtype(data.index), 'data index must be features in net'
    assert isinstance(net, pd.DataFrame), \
    f'net must be a pd.DataFrame containing the columns {x} and {y}'
    assert isinstance(name, str), 'name must be a str'
    assert isinstance(top, int) and top > 0, 'top must be int and > 0'
    assert isinstance(thr_x, (int, float)), 'thr_x must be numeric'
    assert isinstance(thr_y, (int, float)), 'thr_y must be numeric'
    if max_x is None:
        max_x = np.inf
    if max_y is None:
        max_y = np.inf
    assert isinstance(max_x, (int, float)) and max_x > 0, \
    'max_x must be None, or numeric and > 0'
    assert isinstance(max_y, (int, float)) and max_y > 0, \
    'max_y must be None, or numeric and > 0'
    assert isinstance(color_pos, str), 'color_pos must be str'
    assert isinstance(color_neg, str), 'color_neg must be str'
    # Instance
    bp = Plotter(**kwargs)
    # Extract df
    df = data.copy().reset_index(names='target')
    # Filter by net shared targets
    vnet = _validate_net(net)
    snet = vnet[vnet['source'] == name]
    df = pd.merge(df, snet, on=['target'], how='inner').set_index('target')
    # Filter by limits
    msk_x = np.abs(df[x]) < np.abs(max_x)
    msk_y = np.abs(df[y]) < np.abs(max_y)
    df = df.loc[msk_x & msk_y]
    # Define +/- color
    pos = ((df[x] >= 0) & (df[y] >= 0)) | ((df[x] < 0) & (df[y] < 0))
    df['color'] = color_neg
    df.loc[pos, 'color'] = color_pos
    # Plot
    df.plot.scatter(x=x, y=y, c='color', ax=bp.ax)
    # Draw thr lines
    bp.ax.axvline(x=thr_x, linestyle='--', color="black")
    bp.ax.axhline(y=thr_y, linestyle='--', color="black")
    # Add labels
    bp.ax.set_title(name)
    bp.ax.set_xlabel(x)
    bp.ax.set_ylabel(y)
    # Show top features
    df['order'] = df[x].abs() * df[y].abs()
    signs = df.sort_values('order', ascending=False)
    signs = signs.iloc[:top]
    texts = []
    for x, y, s in zip(signs[x], signs[y], signs.index):
        texts.append(bp.ax.text(x, y, s))
    if len(texts) > 0:
        at.adjust_text(texts, arrowprops=dict(arrowstyle='-', color='black'), ax=bp.ax)
    return bp._return()



from typing import Union, Tuple

from anndata import AnnData
import pandas as pd
import numpy as np


DataType = Union[
    AnnData,
    pd.DataFrame,
    Tuple[np.ndarray, np.ndarray, np.ndarray],
]



import logging

import numpy as np
import scipy.stats as sts
import pytest

import decoupler as dc


@pytest.mark.parametrize(
    'nvar,val,size,hasval',
    [
        [3, 0., 5, False],
        [10, 0., 10, True],
    ]
)
def test_fillval(
    nvar,
    val,
    size,
    hasval,
):
    arr = np.array([1., 2., 3., 4., 5.])
    farr = dc.ds._toy._fillval(arr=arr, nvar=nvar, val=val)

    assert farr.size == size
    assert (val == farr[-1]) == hasval


@pytest.mark.parametrize(
    'nobs,nvar,bval,pstime,seed,verbose',
    [
        [10, 15, 2, True, 42, False],
        [2, 12, 2, False, 42, False],
        [100, 50, 0, False, 0, True],
        [10, 500, 0, True, 0, True],
        
    ]
)
def test_toy(
    nobs,
    nvar,
    bval,
    pstime,
    seed,
    verbose,
    caplog,
):
    with caplog.at_level(logging.INFO):
        adata, net = dc.ds.toy(nobs=nobs, nvar=nvar, bval=bval, pstime=pstime, seed=seed, verbose=verbose)
    if verbose:
        assert len(caplog.text) > 0
    else:
        assert caplog.text == ''
    assert all(adata.obs['group'].cat.categories == ['A', 'B'])
    msk = adata.obs['group'] == 'A'
    assert all(adata[msk, :4].X.mean(0) > adata[~msk, :4].X.mean(0))
    assert all(adata[msk, 4:8].X.mean(0) < adata[~msk, 4:8].X.mean(0))
    assert nobs == adata.n_obs
    assert nvar == adata.n_vars
    assert ((bval - 1) < np.mean(adata.X[:, -1].ravel()) < (bval + 1)) or nvar == 12
    if pstime:
        assert 'pstime' in adata.obs.columns
        assert ((0. <= adata.obs['pstime']) & (adata.obs['pstime'] <= 1.)).all()


@pytest.mark.parametrize(
    'shuffle_r,seed,nobs,nvar,is_diff',
    [
        [0.0, 1, 20, 31, True],
        [0.1, 2, 36, 41, True],
        [0.9, 3, 49, 21, False],
        [1.0, 4, 18, 41, False],
        
    ]
)
def test_toy_bench(
    net,
    shuffle_r,
    seed,
    nobs,
    nvar,
    is_diff,
):
    adata, bmnet = dc.ds.toy_bench(shuffle_r=shuffle_r, seed=seed, nobs=nobs, nvar=nvar)
    assert (net == bmnet).values.all()
    assert adata.n_obs == nobs
    assert adata.n_vars == nvar
    msk = adata.obs['group'] == 'A'
    a_adata = adata[msk, :].copy()
    b_adata = adata[~msk, :].copy()
    for j in adata.var_names[:8]:
        a = a_adata[:, j].X.ravel()
        b = b_adata[:, j].X.ravel()
        stat, pval = sts.ranksums(a, b)
        if is_diff:
            assert pval < 0.05
        else:
            assert pval > 0.05

    


import os
import io
import gzip
import warnings
import requests

import pandas as pd
from tqdm.auto import tqdm
import scipy.io as sio
import anndata as ad

from decoupler._docs import docs
from decoupler._log import _log
from decoupler.ds._utils import ensmbl_to_symbol


def _download_anndata(
    url: str,
    verbose: bool = False,
) -> ad.AnnData:
    warnings.filterwarnings("ignore", category=FutureWarning)
    with requests.get(url, stream=True) as response:
        response.raise_for_status()
        total_size = int(response.headers.get('content-length', 0))
        chunk_size = 8192
        buffer = io.BytesIO()
        with tqdm(total=total_size, unit='B', unit_scale=True,
                  desc="Downloading .h5ad", disable=not verbose) as pbar:
            for chunk in response.iter_content(chunk_size=chunk_size):
                buffer.write(chunk)
                pbar.update(len(chunk))
    buffer.seek(0)
    adata = ad.read_h5ad(buffer)
    return adata


@docs.dedent
def pbmc3k(
    verbose: bool = False,
) -> ad.AnnData:
    """
    Downloads single-cell RNA-seq data of peripheral blood mononuclear
    cells (PBMCs) from a healthy donor.

    Parameters
    ----------
    %(verbose)s

    Returns
    -------
    AnnData object.
    """
    url = ('https://raw.githubusercontent.com/chanzuckerberg/' \
    'cellxgene/main/example-dataset/pbmc3k.h5ad')
    adata = _download_anndata(url, verbose=verbose)
    adata = adata.raw.to_adata()
    adata.obs['celltype'] = adata.obs['louvain']
    adata.obs['leiden'] = adata.obs['louvain'].cat.codes
    adata.obs['leiden'] = adata.obs['leiden'].apply(lambda x: f'Clust:{x}')
    adata.obs['leiden'] = adata.obs['leiden'].astype('category')
    del adata.obs['louvain']
    m = f'generated AnnData with shape={adata.shape}'
    _log(m, level='info', verbose=verbose)
    return adata


@docs.dedent
def covid5k(
    verbose: bool = False,
) -> ad.AnnData:
    """
    Downloads single-cell RNA-seq data of peripheral blood mononuclear
    cells (PBMCs) from a cohort of patients with and without
    COVID-19 :cite:`covid5k`.

    Parameters
    ----------
    %(verbose)s

    Returns
    -------
    AnnData object.
    """
    url = (
        'https://ftp.ebi.ac.uk/pub/databases/microarray/data/atlas/'
        'sc_experiments/E-MTAB-9221/E-MTAB-9221.aggregated_counts.mtx.gz'
    )
    url_base = os.path.dirname(url)
    id_ebi, _, _, _ = os.path.basename(url).split('.')
    # Download X
    response = requests.get(url)
    with gzip.GzipFile(fileobj=io.BytesIO(response.content)) as f:
        X = sio.mmread(f).T.tocsr().rint()
        X.eliminate_zeros()
    # Download var
    url_var = os.path.join(url_base, f'{id_ebi}.aggregated_counts.mtx_rows.gz')
    var = pd.read_csv(url_var, sep='\t', header=None, usecols=[0], index_col=0)
    var['name'] = ensmbl_to_symbol(var.index)
    msk_var = ~(var['name'].isna() | var['name'].duplicated(keep='first')).values
    var = var.loc[msk_var].reset_index(drop=True).set_index('name')
    var.index.name = None
    X = X[:, msk_var]
    # Download obs
    url_obs = os.path.join(url_base, f'{id_ebi}.cell_metadata.tsv')
    cols = ['id', 'individual', 'sex', 'disease', 'inferred_cell_type_-_ontology_labels']
    obs = pd.read_csv(url_obs, sep='\t')[cols].set_index('id')
    obs = obs.rename(columns={'inferred_cell_type_-_ontology_labels': 'celltype'})
    obs.index.name = None
    msk_obs = (~obs['celltype'].isna()).values
    obs = obs.loc[msk_obs]
    X = X[msk_obs, :]
    # Make AnnData
    adata = ad.AnnData(X, obs=obs, var=var)
    # Download umap
    url_umap = os.path.join(url_base, f'{id_ebi}.umap_neigh_15.tsv')
    umap = pd.read_csv(url_umap, sep='\t', header=None, index_col=0)
    inter = umap.index.intersection(adata.obs_names)
    adata = adata[inter, :].copy()
    adata.obsm['X_umap'] = umap.loc[inter].values
    # Basic filtering
    msk_var = adata.X.getnnz(axis=0) > 3
    msk_obs = adata.X.getnnz(axis=1) > 3
    adata = adata[msk_obs, msk_var].copy()
    # Make categorical
    for col in adata.obs.columns:
        adata.obs[col] = adata.obs[col].astype('category')
    m = f'generated AnnData with shape={adata.shape}'
    _log(m, level='info', verbose=verbose)
    return adata


@docs.dedent
def erygast1k(
    verbose: bool = False,
) -> ad.AnnData:
    """
    Downloads single-cell RNA-seq data of the erythroid lineage
    during gastrulation in mouse :cite:`erygast`.

    Parameters
    ----------
    %(verbose)s

    Returns
    -------
    AnnData object.
    """
    # How to process from scvelo:
    """
    url = 'https://ndownloader.figshare.com/files/27686871'
    adata = dc.ds._scell._download_anndata(url, verbose=True)
    adata.var = adata.var.drop(columns=['MURK_gene', 'Δm', 'scaled Δm'])
    del adata.layers
    rng = np.random.default_rng(seed=42)
    idx = rng.choice(adata.obs_names, 801, replace=False)
    adata = adata[idx, :].copy()
    msk_var = adata.X.getnnz(axis=0) >= 25
    adata = adata[:, msk_var].copy()
    sc.pp.normalize_total(adata, target_sum=1e4)
    sc.pp.log1p(adata)
    del adata.var
    adata.write('adata.h5ad')
    """
    # Download
    url = "https://zenodo.org/records/15462498/files/adata.h5ad?download=1"
    adata = _download_anndata(url, verbose=verbose)
    m = f'generated AnnData with shape={adata.shape}'
    _log(m, level='info', verbose=verbose)
    return adata



from decoupler.op._translate import show_organisms, translate
from decoupler.op._resource import show_resources, resource
from decoupler.op._collectri import collectri
from decoupler.op._dorothea import dorothea
from decoupler.op._hallmark import hallmark
from decoupler.op._progeny import progeny



from typing import Tuple

import numpy as np
import scipy.stats as sts
import scipy.sparse as sps
from tqdm.auto import tqdm
import numba as nb

from decoupler._docs import docs
from decoupler._log import _log
from decoupler._Method import MethodMeta, Method
from decoupler.pp.net import _getset


@nb.njit(cache=True)
def _std(
    arr: np.ndarray,
    ddof: int,
) -> float:
    N = arr.shape[0]
    m = np.mean(arr)
    var = np.sum((arr - m)**2) / (N - ddof)
    sd = np.sqrt(var)
    return sd


def _ridx(
    times: int,
    nvar: int,
    seed: int | None,
):
    idx = np.tile(np.arange(nvar), (times, 1))
    if seed:
        rng = np.random.default_rng(seed=seed)
        for i in idx:
            rng.shuffle(i)
    return idx


@nb.njit(cache=True)
def _esrank(
    row: np.ndarray,
    rnks: np.ndarray,
    set_msk: np.ndarray,
    dec: float,
) -> Tuple[float, int, np.ndarray]:
    # Init empty
    mx_value = 0.0
    cum_sum = 0.0
    mx_pos = 0.0
    mx_neg = 0.0
    j_pos = 0
    j_neg = 0
    es = np.zeros(rnks.size)
    # Compute norm
    sum_set = np.sum(np.abs(row[set_msk]))
    if sum_set == 0.:
        return 0., 0, np.zeros(rnks.size)
    # Compute ES
    for i in rnks:
        if set_msk[i]:
            cum_sum += np.abs(row[i]) / sum_set
            es[i] = cum_sum
        else:
            cum_sum -= dec
            es[i] = cum_sum
        # Update max scores and idx
        if cum_sum > mx_pos:
            mx_pos = cum_sum
            j_pos = i
        if cum_sum < mx_neg:
            mx_neg = cum_sum
            j_neg = i
    # Determine if pos or neg are more enriched
    if mx_pos > -mx_neg:
        mx_value = mx_pos
        j = j_pos
    else:
        mx_value = mx_neg
        j = j_neg
    return mx_value, j, es


@nb.njit(cache=True)
def _nesrank(
    ridx: np.ndarray,
    row: np.ndarray,
    rnks: np.ndarray,
    set_msk: np.ndarray,
    dec: float,
    es: float,
) -> Tuple[float, float]:
    # Keep old set_msk upstream
    set_msk = set_msk.copy()
    # Compute null
    times, nvar = ridx.shape
    if times == 0:
        return 0., 1.
    null = np.zeros(times)
    for i in range(times):
        null[i], _, _ = _esrank(row=row, rnks=rnks, set_msk=set_msk[ridx[i]], dec=dec)
    # Compute NES
    pos_null_msk = null >= 0.
    neg_null_msk = null < 0.
    pos_null_sum = pos_null_msk.sum()
    neg_null_sum = neg_null_msk.sum()
    if (es >= 0) and (pos_null_sum > 0):
        pval = (null[pos_null_msk] >= es).sum() / pos_null_sum
        pos_null_mean = null[pos_null_msk].mean()
        nes = es / pos_null_mean
    elif (es < 0) and (neg_null_sum > 0):
        pval = (null[neg_null_msk] <= es).sum() / neg_null_sum
        neg_null_mean = null[neg_null_msk].mean()
        nes = -es / neg_null_mean
    else:
        nes = np.inf
        pval = np.inf
    return nes, pval


@nb.njit(parallel=True, cache=True)
def _stsgsea(
    row: np.ndarray,
    cnct: np.ndarray,
    starts: np.ndarray,
    offsets: np.ndarray,
    ridx: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    # Sort features
    idx = np.argsort(-row)
    row = row[idx]
    # Init empty
    nvar = row.size
    nsrc = starts.size
    rnks = np.arange(nvar)
    es = np.zeros(nsrc)
    nes = np.zeros(nsrc)
    pv = np.ones(nsrc)
    for j in nb.prange(nsrc):
        # Extract fset
        fset = _getset(cnct, starts, offsets, j)
        # Get decending penalty
        dec = 1.0 / (nvar - fset.size)
        # Get msk
        set_msk = np.zeros(nvar, dtype=nb.bool_)
        set_msk[fset] = True
        set_msk = set_msk[idx]
        # Compute es per feature
        es[j], _, _ = _esrank(row=row, rnks=rnks, set_msk=set_msk, dec=dec)
        nes[j], pv[j] = _nesrank(ridx=ridx, row=row, rnks=rnks, set_msk=set_msk, dec=dec, es=es[j])
    return es, nes, pv


@docs.dedent
def _func_gsea(
    mat: np.ndarray,
    cnct: np.ndarray,
    starts: np.ndarray,
    offsets: np.ndarray,
    times: int | float = 1000,
    seed: int | float = 42,
    verbose: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    r"""
    Gene Set Enrichment Analysis (GSEA) :cite:`gsea`.

    Features are ranked based on a continuous statistic (e.g., expression, score, or correlation).
    The enrichment score (ES) for a feature set is computed by walking down the ranked list and increasing a running-sum
    statistic when a feature is in the set, and decreasing it when it is not.

    .. math::
    
       \delta(F, i) = 
       \begin{cases}
       \frac{|r_i|}{\sum\limits_{j \in F} |r_j|} & \text{if feature } i \in F \\
       -\frac{1}{l} & \text{if feature } i \notin F
       \end{cases}

    Where:

    - :math:`F` is a feature set
    - :math:`r` is the ranking of the feature statistics in descending order
    - :math:`r_i` is the value for feature :math:`i`
    - :math:`r_j` is the value for feature :math:`j` in :math:`F`
    - :math:`k` is the number of features in :math:`F`
    - :math:`N` is the total number of features in :math:`r`
    - :math:`l=N-k` is the number of features not in :math:`F` but present in :math:`r`

    For each feature, the function :math:`\delta(F,i)` is applied and stored as a sequence :math:`L`.

    .. math::

        L = \delta(F, i)\text{ for i} = \text{1, 2, ... , N}

    The enrichment score :math:`ES` corresponds to the maximum deviation from zero of this running sum.

    .. math::

        ES = L_{arg max |L|}

    When multiple random permutations are done, statistical significance is assessed via empirical testing.

    .. math::

        p_{value}=\frac{ES_{rand} \geq ES}{P}

    Where:

    - :math:`ES_{rand}` are the enrichment scores of the random permutations
    - :math:`P` is the total number of permutations

    Additionaly, :math:`ES` is updated to a normalized enrichment score :math:`NES`.

    .. math::

        NES = \begin{cases} \frac{ES}{\mu_{+}} & \text{if } ES > 0 \\ \frac{ES}{\mu_{-}} & \text{if } ES < 0  \end{cases}

    Where:

    - :math:`\mu{+}` is the mean of positive values in :math:`ES_{rand}`
    - :math:`\mu{-}` is the mean of negative values in :math:`ES_{rand}`

    Finally, the obtained math:`p_value` are adjusted by Benjamini-Hochberg correction. 

    %(params)s
    %(times)s
    %(seed)s

    %(returns)s
    """
    nobs, nvar = mat.shape
    assert isinstance(times, (int, float)) and times >= 0, 'times must be numeric and >= 0'
    assert isinstance(seed, (int, float)) and seed >= 0, 'seed must be numeric and >= 0'
    times, seed = int(times), int(seed)
    # Compute
    nsrc = starts.size
    m = f'gsea - calculating {nsrc} scores across {nobs} observations'
    _log(m, level='info', verbose=verbose)
    if times > 1:
        m = f'gsea - comparing estimates against {times} random permutations'
        _log(m, level='info', verbose=verbose)
        ridx = _ridx(times=times, nvar=nvar, seed=seed)
    else:
        ridx = _ridx(times=times, nvar=nvar, seed=None)
    es = np.zeros(shape=(nobs, nsrc))
    nes = np.zeros(shape=(nobs, nsrc))
    pv = np.zeros(shape=(nobs, nsrc))
    for i in tqdm(range(nobs), disable=not verbose):
        if isinstance(mat, sps.csr_matrix):
            row = mat[i].toarray()[0]
        else:
            row = mat[i]
        es[i, :], nes[i, :], pv[i, :] = _stsgsea(
            row=row,
            cnct=cnct,
            starts=starts,
            offsets=offsets,
            ridx=ridx,
        )
    if times > 1:
        es = nes
    return es, pv


_gsea = MethodMeta(
    name='gsea',
    desc='Gene Set Enrichment Analysis (GSEA)',
    func=_func_gsea,
    stype='numerical',
    adj=False,
    weight=False,
    test=True,
    limits=(-np.inf, +np.inf),
    reference='https://doi.org/10.1073/pnas.0506580102',
)
gsea = Method(_method=_gsea)



from typing import Tuple

import numpy as np
import scipy.stats as sts
import scipy.sparse as sps
from tqdm.auto import tqdm
import numba as nb

from decoupler._docs import docs
from decoupler._log import _log
from decoupler._Method import MethodMeta, Method
from decoupler.pp.net import _getset


@nb.njit(parallel=True, cache=True)
def _auc(
    row: np.ndarray,
    cnct: np.ndarray,
    starts: np.ndarray,
    offsets: np.ndarray,
    n_up: int,
    nsrc: int,
) -> np.ndarray:
    # Empty acts
    es = np.zeros(nsrc)
    # For each feature set
    for j in nb.prange(nsrc):
        # Extract feature set
        fset = _getset(cnct, starts, offsets, j)
        # Compute max AUC for fset
        x_th = np.arange(1, stop=fset.shape[0] + 1)
        x_th = x_th[x_th < n_up]
        max_auc = np.sum(np.diff(np.append(x_th, n_up)) * x_th)
        # Compute AUC
        x = row[fset]
        x = np.sort(x[x <= n_up])
        y = np.arange(x.shape[0]) + 1
        x = np.append(x, n_up)
        # Update acts matrix
        es[j] = np.sum(np.diff(x) * y) / max_auc
    return es


def _validate_n_up(
    nvar: int,
    n_up: int | float | None = None,
) -> int:
    assert isinstance(n_up, (int, float)) or n_up is None, 'n_up must be numerical or None'
    if n_up is None:
        n_up = np.ceil(0.05 * nvar)
        n_up = int(np.clip(n_up, a_min=2, a_max=nvar))
    else:
        n_up = int(np.ceil(n_up))
    assert nvar >= n_up > 1, f'For nvar={nvar}, n_up={n_up} must be between 1 and {nvar}'
    return n_up


@docs.dedent
def _func_aucell(
    mat: np.ndarray,
    cnct: np.ndarray,
    starts: np.ndarray,
    offsets: np.ndarray,
    n_up: int | float | None = None,
    verbose: bool = False,
) -> Tuple[np.ndarray, None]:
    """
    Area Under the Curve for set enrichment within single cells (AUCell) :cite:`aucell`.

    Given a ranked list of features per observation, AUCell calculates the AUC by measuring how early the features in
    the set appear in this ranking. Specifically, the enrichment score :math:`ES` is:

    .. math::
    
       {ES}_{i, F} = \int_0^1 {RecoveryCurve}_{i, F}(r_i) \, dr
    
    Where:
    
    - :math:`i` is the obervation  
    - :math:`F` is the feature set  
    - :math:`{RecoveryCurve}_{i, F}(r_i)` is the proportion of features from :math:`F` recovered in the top :math:`r_i`-fraction of the ranked list for observation :math:`i`

    %(notest)s

    %(params)s
    n_up
        Number of features to include in the AUC calculation.
        If ``None``, the top 5% of features based on their magnitude are selected.

    %(returns)s
    """
    nobs, nvar = mat.shape
    nsrc = starts.size
    n_up = _validate_n_up(nvar, n_up)
    m = f'aucell - calculating {nsrc} AUCs for {nvar} targets across {nobs} observations, categorizing features at rank={n_up}' 
    _log(m, level='info', verbose=verbose)
    es = np.zeros(shape=(nobs, nsrc))
    for i in tqdm(range(mat.shape[0]), disable=not verbose):
        if isinstance(mat, sps.csr_matrix):
            row = mat[i].toarray()[0]
        else:
            row = mat[i]
        row = sts.rankdata(a=-row, method='ordinal')
        es[i] = _auc(row=row, cnct=cnct, starts=starts, offsets=offsets, n_up=n_up, nsrc=nsrc)
    return es, None


_aucell = MethodMeta(
    name='aucell',
    desc='AUCell',
    func=_func_aucell,
    stype='categorical',
    adj=False,
    weight=False,
    test=False,
    limits=(0, 1),
    reference='https://doi.org/10.1038/nmeth.4463',
)
aucell = Method(_method=_aucell)



from typing import Tuple

import pandas as pd
import numpy as np
import scipy.sparse as sps
from anndata import AnnData

from decoupler._log import _log
from decoupler.pp.net import prune


def _validate_groupby(
    obs: pd.DataFrame,
    groupby: str | list | None,
    runby: str,
) -> None | list:
    assert isinstance(groupby, (str, list)) or groupby is None, \
    'groupby must be str, list or None'
    assert isinstance(runby, str) and runby in ['expr', 'source'], \
    'runby must be str and either expr or source'
    if groupby is not None:
        if type(groupby) is str:
            groupby = [groupby]
        for grp_i in groupby:
            if type(grp_i) is str:
                grp_i = [grp_i]
            # For each group inside each groupby
            for grp_j in grp_i:
                assert not ('source' == grp_j and runby == 'source'), \
                f'source cannot be in groupby if runby="source"'
                # Assert that columns exist in obs
                assert grp_j in obs.columns, \
                f'Column name "{grp_j}" must be in adata.obs.columns'
                # Assert that column doesn't contain "|"
                assert '|' not in grp_j, \
                'Column names must not contain the \"|\" character'
        return groupby


def _validate_obs(
    obs: pd.DataFrame,
) -> None:
    assert 'source' in obs.columns, \
    'source must be in adata.obs.columns'
    assert 'type_p' in obs.columns, \
    'type_p must be in adata.obs.columns'
    assert pd.api.types.is_numeric_dtype(obs['type_p']), \
    'type_p must contain numeric values'
    assert np.isin(obs['type_p'].sort_values().unique(), np.array([-1, 1])).all(), \
    'type_p must be -1 or +1'


def _filter(
    adata: AnnData,
    net: pd.DataFrame,
    sfilt: bool,
    verbose: bool,
) -> Tuple[AnnData, pd.DataFrame]:
    # Remove experiments without sources in net
    srcs = net['source'].unique()
    prts = set()
    msk_exp = np.zeros(adata.obs_names.size, dtype=np.bool_)
    for i, src in enumerate(adata.obs['source']):
        if isinstance(src, list):
            prts.update(src)
            if np.isin(src, srcs).any():
                msk_exp[i] = True
        elif isinstance(src, str):
            prts.add(src)
            if src in srcs:
                msk_exp[i] = True
    n_exp = adata.shape[0]
    m = f'benchmark - found {len(prts)} unique perturbed sources across {n_exp} experiments'
    _log(m, level='info', verbose=verbose)
    r_exp = int((~msk_exp).sum())
    m = f'benchmark - removing {r_exp} experiments out of {n_exp} without sources in net'
    _log(m, level='info', verbose=verbose)
    adata = adata[msk_exp, :].copy()
    # Remove sources without experiments in obs
    if sfilt:
        msk_src = np.array([s in prts for s in net['source']])
        rsrc = net.loc[~msk_src].groupby('source').size().index.size
        m = f'benchmark - removing {rsrc} sources out of {srcs.size} without experiments in obs'
        _log(m, level='info', verbose=verbose)
        net = net.loc[msk_src, :]
    adata.uns['p_sources'] = prts
    return adata, net


def _sign(
    adata: AnnData,
) -> None:
    v_sign = adata.obs['type_p'].values.reshape(-1, 1)
    if sps.issparse(adata.X):
        adata.layers['tmp'] = adata.X.multiply(v_sign).tocsr()
    else:
        adata.layers['tmp'] = adata.X * v_sign


def _validate_bool(
    y_true: np.ndarray,
    y_score: np.ndarray,
) -> None:
    assert isinstance(y_true, np.ndarray), 'y_true must be numpy.ndarray'
    assert isinstance(y_score, np.ndarray), 'y_score must be numpy.ndarray'
    unq = np.sort(np.unique(y_true))
    m = 'y_true must contain two binary classes, 0 and 1'
    assert unq.size <= 2, m
    lbl = np.array([0, 1])
    assert np.all(unq == lbl), m
    assert y_true.size == y_score.size, \
    'y_true and y_score must have the same size'



from typing import Tuple, Callable
import inspect
from functools import partial

import numpy as np
import scipy.stats as sts
import numba as nb

from decoupler._docs import docs
from decoupler._log import _log
from decoupler._Method import MethodMeta, Method
from decoupler.mt._gsea import _std, _ridx


@nb.njit(cache=True)
def _wsum(
    x: np.ndarray,
    w: np.ndarray,
) -> float:
    return np.sum(x * w)


@nb.njit(cache=True)
def _wmean(
    x: np.ndarray,
    w: np.ndarray,
) -> float:
    agg = _wsum(x, w)
    div = np.sum(np.abs(w))
    return agg / div


def _fun(
    f: Callable,
    verbose: bool = False,
):
    @nb.njit(parallel=True, cache=True)
    def _f(mat, adj):
        nobs, nvar = mat.shape
        nvar, nsrc = adj.shape
        es = np.zeros((nobs, nsrc))
        for i in nb.prange(nobs):
            x = mat[i]
            for j in range(nsrc):
                w = adj[:, j]
                es[i, j] = f(x, w)
        return es
    _f.__name__ = f.__name__
    if _f.__name__ not in _cfuncs:
        _cfuncs[f.__name__] = _f
        m = f'waggr - using {_f.__name__} for the first time, will need to be compiled'
        _log(m, level='info', verbose=verbose)


_fun_dict = {
    'wsum': _wsum,
    'wmean': _wmean,
}

_cfuncs = dict()

def _validate_args(
    fun: Callable,
    verbose: bool,
) -> bool:
    args = inspect.signature(fun).parameters
    required_args = ['x', 'w']
    for arg in required_args:
        if arg not in args:
            assert False, f'fun={fun.__name__} must contain arguments x and w'
    # Check if any additional arguments have default values
    for param in args.values():
        if param.name not in required_args and param.default == inspect.Parameter.empty:
            assert False, f'fun={fun.__name__} has an argument {param.name} without a default value'
    if not hasattr(fun, 'func_code'):
        m = f'waggr - {fun.__name__} will be compiled into numba code'
        _log(m, level='info', verbose=verbose)
        fun = nb.njit(fun)
    return fun


def _validate_func(
    fun: str | Callable,
    verbose: bool,
) -> None:
    fun = _validate_args(fun=fun, verbose=verbose)
    x = np.array([1., 2., 3.])
    w = np.array([-1., 0., 2.])
    try:
        res = fun(x=x, w=w)
        assert isinstance(res, (int, float)), 'output of fun must be a single numerical value'
    except:
        raise ValueError(f'fun failed to run with test data: fun(x={x}), w={w}')
    m = f'waggr - using function {fun.__name__}'
    _log(m, level='info', verbose=verbose)
    _fun(f=fun, verbose=verbose)


@nb.njit(parallel=True, cache=True)
def _perm(
    fun: Callable,
    es: np.ndarray,
    mat: np.ndarray,
    adj: np.ndarray,
    idx: np.ndarray,
):
    # Init
    nobs, nvar = mat.shape
    nvar, nsrc = adj.shape
    times, nvar = idx.shape
    null_dst = np.zeros((nobs, nsrc, times))
    pvals = np.zeros((nobs, nsrc))
    # Permute
    for i in nb.prange(times):
        null_dst[:, :, i] = fun(mat[:, idx[i]], adj)
        pvals += np.abs(null_dst[:, :, i]) > np.abs(es)
    # Compute z-score
    nes = np.zeros(es.shape)
    for i in nb.prange(nobs):
        for j in range(nsrc):
            e = es[i, j]
            d = _std(null_dst[i, j, :], 1)
            if d != 0.:
                n = (e - np.mean(null_dst[i, j, :])) / d
            else:
                if e != 0.:
                    n = np.inf
                else:
                    n = 0.
            nes[i, j] = n
    # Compute empirical p-value
    pvals = np.where(pvals == 0.0, 1.0, pvals)
    pvals = np.where(pvals == times, times - 1, pvals)
    pvals = pvals / times
    pvals = np.where(pvals >= 0.5, 1 - (pvals), pvals)
    pvals = pvals * 2
    return nes, pvals


@docs.dedent
def _func_waggr(
    mat: np.ndarray,
    adj: np.ndarray,
    fun: str | Callable = 'wmean',
    times: int | float = 1000,
    seed: int | float = 42,
    verbose: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Weighted Aggregate (WAGGR) :cite:`decoupler`.

    %(yestest)s

    %(params)s
    
    fun
        Function to compute enrichment statistic from omics readouts (``x``) and feature weights (``w``).
        Provided function must contain ``x`` and ``w`` arguments and ouput a single float.
        By default, 'wmean' and 'wsum' are implemented.
    %(times)s
    %(seed)s

    %(returns)s
    """
    assert isinstance(fun, str) or callable(fun), 'fun must be str or callable'
    if isinstance(fun, str):
        assert fun in _fun_dict, 'when fun is str, it must be wmean or wsum'
        fun = _fun_dict[fun]
    _validate_func(fun, verbose=verbose)
    vfun = _cfuncs[fun.__name__]
    assert isinstance(times, (int, float)) and times >= 0, 'times must be numeric and >= 0'
    assert isinstance(seed, (int, float)) and seed >= 0, 'seed must be numeric and >= 0'
    times, seed = int(times), int(seed)
    nobs, nvar = mat.shape
    nvar, nsrc = adj.shape
    m = f'waggr - calculating scores for {nsrc} sources across {nobs} observations'
    _log(m, level='info', verbose=verbose)
    es = vfun(mat, adj)
    if times > 1:
        m = f'waggr - comparing estimates against {times} random permutations'
        _log(m, level='info', verbose=verbose)
        idx = _ridx(times=times, nvar=nvar, seed=seed)
        es, pv = _perm(fun=vfun, es=es, mat=mat, adj=adj, idx=idx)
    else:
        pv = np.ones(es.shape)
    return es, pv


_waggr = MethodMeta(
    name='waggr',
    desc='Weighted Aggregate (WAGGR)',
    func=_func_waggr,
    stype='numerical',
    adj=True,
    weight=True,
    test=True,
    limits=(-np.inf, +np.inf),
    reference='https://doi.org/10.1093/bioadv/vbac016',
)
waggr = Method(_method=_waggr)



import pandas as pd
import matplotlib
matplotlib.use("Agg")
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import pytest

import decoupler as dc


@pytest.fixture
def tdata_obsm_pca(
    tdata_obsm,
):
    dc.tl.rankby_obsm(tdata_obsm, key='X_pca')
    return tdata_obsm


@pytest.fixture
def tdata_obsm_ulm(
    tdata_obsm,
):
    tdata_obsm = tdata_obsm.copy()
    dc.tl.rankby_obsm(tdata_obsm, key='score_ulm')
    return tdata_obsm


@pytest.mark.parametrize(
    'pca,names,nvar,dendrogram,titles,cmap_obs',
    [
        [True, None, 10, True, ['Scores', 'Stats'], dict()],
        [True, 'group', 5, False, ['asd', 'fgh'], dict()],
        [True, ['group', 'pstime'], 10, True, ['Scores', 'Stats'], dict()],
        [True, None, 10, True, ['Scores', 'Stats'], dict(group='tab10', pstime='magma', sample='Pastel1')],
        [True, None, 2, True, ['Scores', 'Stats'], dict(pstime='magma')],
        [True, None, ['PC01', 'PC02'], True, ['Scores', 'Stats'], dict(pstime='magma')],
        [False, None, None, True, ['Scores', 'Stats'], dict()],
        [False, None, 10, True, ['Scores', 'Stats'], dict()],
        [False, None, 'T3', True, ['Scores', 'Stats'], dict()],
        [False, None, ['T5', 'T3'], True, ['Scores', 'Stats'], dict()],
    ]
)
def test_obsm(
    tdata_obsm_pca,
    tdata_obsm_ulm,
    pca,
    names,
    nvar,
    dendrogram,
    titles,
    cmap_obs,
):
    if pca:
        tdata_obsm_ranked = tdata_obsm_pca
    else:
        tdata_obsm_ranked = tdata_obsm_ulm
    fig = dc.pl.obsm(
        tdata_obsm_ranked,
        names=names,
        nvar=nvar,
        dendrogram=dendrogram,
        titles=titles,
        cmap_obs=cmap_obs,
        return_fig=True
    )
    assert isinstance(fig, Figure)
    plt.close(fig)



from .net import read_gmt, prune, adjmat, idxmat, shuffle_net, net_corr
from .data import extract
from .anndata import get_obsm, swap_layer, pseudobulk, filter_samples, \
filter_by_expr, filter_by_prop, knn, bin_order



import pandas as pd
import seaborn as sns
from matplotlib.figure import Figure

from decoupler._docs import docs
from decoupler._Plotter import Plotter
from decoupler.bm.pl._format import _format


@docs.dedent
def bar(
    df: pd.DataFrame,
    x: str,
    y: str,
    hue: str | None = None,
    palette: str = 'tab20',
    **kwargs
) -> None | Figure:
    """
    Plot the harmonic mean between two metric statistics as a barplot.

    x-axis represent the harmonic mean between metric statistics.

    y-axis represent a grouping variable.

    Parameters
    ----------
    %(df)s
    x
        Continous variable to plot on x axis.
    %(y)s
    %(hue)s
    %(palette)s
    %(plot)s
    """
    # Validate
    assert isinstance(x, str), 'x must be str'
    assert isinstance(y, str), 'y must be str'
    assert isinstance(hue, str) or hue is None, 'hue must be str or None'
    # Instance
    bp = Plotter(**kwargs)
    # Plot
    order = (
        df
        .groupby(y)[x]
        .mean()
        .sort_values(ascending=False)
        .index
    )
    args = dict()
    if hue is not None:
        args['hue'] = hue
        args['palette'] = palette
    sns.barplot(
        data=df,
        y=y,
        x=x,
        order=order,
        **args
    )
    if hue is not None and hue != y:
        bp.ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False, title=hue)
    return bp._return()



import requests
from tqdm import tqdm
import pandas as pd
import io

from decoupler._log import _log

URL_DBS = 'https://omnipathdb.org/annotations?databases='
URL_INT = 'https://omnipathdb.org/interactions/?genesymbols=1&'

def _download(
    url: str,
    verbose: bool = False,
    **kwargs,
) -> pd.DataFrame:
    assert isinstance(url, str), 'url must be str'
    # Download with progress bar
    m = f'Downloading {url}'
    _log(m, level='info', verbose=verbose)
    chunks = []
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with tqdm(unit='B', unit_scale=True, desc="Progress", disable=not verbose) as pbar:
            for chunk in r.iter_content(chunk_size=8192):
                if chunk:
                    chunks.append(chunk)
                    pbar.update(len(chunk))
    # Read into pandas
    data = io.BytesIO(b"".join(chunks))
    df = pd.read_csv(data, **kwargs)
    m = f'Download finished'
    _log(m, level='info', verbose=verbose)
    return df



import requests
import os
import io

import pandas as pd


def ensmbl_to_symbol(
    genes: list,
) -> list:
    """
    Transforms ensembl gene ids to gene symbols.

    Parameters
    ----------
    genes
        List of ensembl gene ids to transform.

    Returns
    -------
    List of gene symbols
    """
    url = (
        'http://www.ensembl.org/biomart/martservice?query=<?xml version="1.0" encoding="UTF-8"?>'
        '<!DOCTYPE Query><Query  virtualSchemaName = "default" formatter = "TSV" header = "0" un'
        'iqueRows = "0" count = "" datasetConfigVersion = "0.6" ><Dataset name = "hsapiens_gene_'
        'ensembl" interface = "default" ><Attribute name = "ensembl_gene_id" /><Attribute name ='
        '"external_gene_name" /></Dataset></Query>'
    )
    # Send a GET request to the URL
    response = requests.get(url)
    eids = pd.read_csv(io.StringIO(response.text), sep='\t', header=None, index_col=0)[1].to_dict()
    return [eids[g] if g in eids else None for g in genes]



import pandas as pd
import numpy as np
import pytest

import decoupler as dc


@pytest.mark.parametrize(
    'levels,dict_weights',
    [
        ['A', None],
        [['A', 'B'], dict(A=1, B=0.5)],
    ]
)
def test_dorothea(
    levels,
    dict_weights,
):
    do = dc.op.dorothea(levels=levels, dict_weights=dict_weights)
    assert isinstance(do, pd.DataFrame)
    cols = {'source', 'target', 'weight', 'confidence'}
    assert cols.issubset(do.columns)
    assert pd.api.types.is_numeric_dtype(do['weight'])
    assert not do.duplicated(['source', 'target']).any()



import pandas as pd
import scipy.stats as sts
import scipy.sparse as sps
from anndata import AnnData
from xgboost import XGBRegressor

from decoupler._docs import docs
from decoupler.pp.data import extract


@docs.dedent
def rankby_order(
    adata: AnnData,
    order: str,
    thr_padj: int | float = 0.05,
    seed: int = 42,
    **kwargs
) -> pd.DataFrame:
    """
    Rank features along a continuous, ordered process such as pseudotime.

    Parameters
    ----------
    %(adata)s
    %(order)s
    thr_padj
        Threshold used to assign significance after FDR correction.
    %(seed)s

    Returns
    -------
    DataFrame with features associated with the ordering variable. For each feature the following statistics are reported:
    - importance of the ``XGBRegressor``
    - Pearson correlation coefficient
    - the sign of the association, 0 if the correltation is non-significant, and +1 or -1 depending on the correlation sign.
    """
    # Get vars and ordinal variable
    X = adata.X
    if sps.issparse(X):
        X = X.toarray()
    y = adata.obs[order].values
    # Fit
    reg = XGBRegressor(random_state=seed, **kwargs).fit(X, y)
    df = pd.DataFrame()
    df['name'] = adata.var_names
    df['impr'] = reg.feature_importances_
    df['corr'], df['pval'] = sts.pearsonr(X, y.reshape(-1, 1), axis=0)
    df['padj'] = sts.false_discovery_control(df['pval'])
    df = df.sort_values('impr', ascending=False).reset_index(drop=True)
    # Find direction of change
    sign = []
    for corr, padj in zip(df['corr'], df['padj']):
        if padj < thr_padj:
            if corr > 0:
                s = 1
            else:
                s = -1
        else:
            s = 0
        sign.append(s)
    df['sign'] = sign
    return df



import warnings

import pandas as pd
import pytest
import anndata as ad

import decoupler as dc


def test_msvisium():
    adata = dc.ds.msvisium()
    assert isinstance(adata, ad.AnnData)
    assert adata.raw is None
    assert isinstance(adata.obs, pd.DataFrame)
    cols = {'niches'}
    assert cols.issubset(adata.obs.columns)
    for col in cols:
        assert isinstance(adata.obs[col].dtype, pd.CategoricalDtype)



import pandas as pd
import numpy as np
import scipy.stats as sts
from anndata import AnnData

from decoupler._docs import docs


@docs.dedent
def rankby_group(
    adata: AnnData,
    groupby: str,
    reference: str | list = 'rest',
    method: str = 't-test_overestim_var',
) -> pd.DataFrame:
    """
    Rank features for characterizing groups.

    Parameters
    ----------
    %(adata)s
    groupby
        The key of the observations grouping to consider.
    reference
        Reference group or list of reference groups to use as reference.
    method
        Statistical method to use for computing differences between groups.
        Avaliable methods include: ``{'wilcoxon', 't-test', 't-test_overestim_var'}``.

    Returns
    -------
    DataFrame with different features between groups.
    """
    assert isinstance(adata, AnnData), 'adata must be anndata.AnnData'
    assert isinstance(groupby, str) and groupby in adata.obs.columns, \
    'groupby must be str and in adata.obs.columns'
    assert isinstance(reference, (str, list)), 'reference must be str or list'
    methods = {'wilcoxon', 't-test', 't-test_overestim_var'}
    assert isinstance(method, str) and method in methods, \
    f'method must be one of: {methods}'
    
    # Get tf names
    features = adata.var_names
    # Generate mask for group samples
    groups = adata.obs[groupby].unique()
    results = []
    for group in groups:
        # Extract group mask
        g_msk = (adata.obs[groupby] == group).values
        # Generate mask for reference samples
        if reference == 'rest':
            ref_msk = ~g_msk
            ref = reference
        elif isinstance(reference, str):
            ref_msk = (adata.obs[groupby] == reference).values
            ref = reference
        else:
            cond_lst = np.array([(adata.obs[groupby] == r).values for r in reference])
            ref_msk = np.sum(cond_lst, axis=0).astype(bool)
            ref = ', '.join(reference)
        assert np.sum(ref_msk) > 0, f'No reference samples found for {reference}'
        # Skip if same than ref
        if group == ref:
            continue
        # Test differences
        result = []
        for i in np.arange(len(features)):
            v_group = adata.X[g_msk, i]
            v_rest = adata.X[ref_msk, i]
            assert np.all(np.isfinite(v_group)) and np.all(np.isfinite(v_rest)), \
                "adata contains not finite values, please remove them."
            if method == 'wilcoxon':
                stat, pval = sts.ranksums(v_group, v_rest)
            elif method == 't-test':
                stat, pval = sts.ttest_ind_from_stats(
                    mean1=np.mean(v_group),
                    std1=np.std(v_group, ddof=1),
                    nobs1=v_group.size,
                    mean2=np.mean(v_rest),
                    std2=np.std(v_rest, ddof=1),
                    nobs2=v_rest.size,
                    equal_var=False,  # Welch's
                )
            elif method == 't-test_overestim_var':
                stat, pval = sts.ttest_ind_from_stats(
                    mean1=np.mean(v_group),
                    std1=np.std(v_group, ddof=1),
                    nobs1=v_group.size,
                    mean2=np.mean(v_rest),
                    std2=np.std(v_rest, ddof=1),
                    nobs2=v_group.size,
                    equal_var=False,  # Welch's
                )
            mc = np.mean(v_group) - np.mean(v_rest)
            result.append([group, ref, features[i], stat, mc, pval])
        # Tranform to df
        result = pd.DataFrame(
            result,
            columns=['group', 'reference', 'name', 'stat', 'meanchange', 'pval']
        )
        # Correct pvalues by FDR
        result['pval'] = result['pval'].fillna(1)
        result['padj'] = sts.false_discovery_control(result['pval'], method='bh')
        # Sort and save
        result['abs_stat'] = result['stat'].abs()
        result = result.sort_values(['padj', 'pval', 'stat'], ascending=[True, True, False])
        result = result.drop(columns=['abs_stat'])
        results.append(result)
    # Merge
    results = pd.concat(results)
    # Convert to categorical
    results['group'] = results['group'].astype('category')
    results['reference'] = results['reference'].astype('category')
    results['name'] = results['name'].astype('category')
    return results.reset_index(drop=True)



import pytest

import decoupler as dc


@pytest.mark.parametrize(
    'methods,args,consensus',
    [
        ['all', dict(), True],
        ['aucell', dict(aucell=dict(n_up=3)), True],
        [['ulm'], dict(), False],
        [['ulm', 'ora'], dict(ulm=dict(), ora=dict(n_up=3)), False]
    ]
)
def test_decouple(
    adata,
    net,
    methods,
    args,
    consensus,
):
    dc.mt.decouple(data=adata, net=net, methods=methods, args=args, cons=consensus, tmin=0)
    if consensus:
        assert 'score_consensus' in adata.obsm
    else:
        assert 'score_consensus' not in adata.obsm



import pandas as pd


def _infer_dtypes(
    df: pd.DataFrame
) -> pd.DataFrame:
    for col in df.columns:
        try:
            df[col] = pd.to_numeric(df[col])
            continue
        except ValueError:
            pass
        if df[col].dtype == 'string':
            df[col] = df[col].astype(str)
        lowered = df[col].str.lower()
        if lowered.isin(["true", "false"]).all():
            df[col] = lowered == "true"
            continue
    return df



from typing import Tuple

import pandas as pd
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.colors import TwoSlopeNorm

from decoupler._Plotter import Plotter
from decoupler.pp.net import prune
from decoupler.mt._gsea import _esrank


def leading_edge(
    df: pd.DataFrame,
    net: pd.DataFrame,
    stat: str,
    name: str,
    cmap='RdBu_r',
    color='#88c544',
    **kwargs
) -> Tuple[None | Figure, np.ndarray]:
    """
    Plot the running score of GSEA.
    """
    class MidpointNormalize(matplotlib.colors.Normalize):
        def __init__(self, vmin=None, vmax=None, vcenter=None, clip=False):
            self.vcenter = vcenter
            super().__init__(vmin, vmax, clip)
        def __call__(self, value, clip=None):
            x, y = [self.vmin, self.vcenter, self.vmax], [0, 0.5, 1]
            return np.ma.masked_array(np.interp(value, x, y))
    # Validate
    assert isinstance(df, pd.DataFrame), 'df must be pd.DataFrame'
    assert isinstance(stat, str) and stat in df.columns, 'stat must be str and in pd.DataFrame'
    assert isinstance(net, pd.DataFrame), 'net must be pd.DataFrame'
    assert isinstance(name, str), 'name must be str'
    # Extract feature level stats and names from df
    c = df.index.values.astype('U')
    m = df[stat].values
    # Remove empty values
    msk = np.isfinite(m)
    c = c[msk]
    m = m[msk]
    # Transform net
    snet = prune(features=c, net=net, tmin=0, verbose=False)
    snet = snet[snet['source'] == name]
    # Sort features
    idx = np.argsort(-m)
    m = m[idx]
    c = c[idx]
    # Get ranks
    rnks = np.arange(c.size)
    # Get msk
    set_msk = np.isin(c, snet['target'])
    # Get decending penalty
    n_features = set_msk.size
    nf_in_set = set_msk.sum()
    dec = 1.0 / (n_features - nf_in_set)
    # Compute es
    mx_value, j, es = _esrank(row=m, rnks=rnks, set_msk=set_msk, dec=dec)
    # Get leading edge features
    sign = np.sign(mx_value)
    set_rnks = rnks[set_msk]
    if sign > 0:
        le_c = c[set_rnks[set_rnks <= j]]
    else:
        le_c = np.flip(c[set_rnks[set_rnks >= j]])
    # Instance
    kwargs['ax'] = None
    bp = Plotter(**kwargs)
    bp.fig.delaxes(bp.ax)
    plt.close(bp.fig)
    # Plot
    gridspec_kw = {'height_ratios': [4, 0.5, 0.5, 2]}
    bp.fig, axes = plt.subplots(4, 1, gridspec_kw=gridspec_kw, figsize=bp.figsize, sharex=True, dpi=bp.dpi)
    axes = axes.ravel()
    # Plot random walk
    ax = axes[0]
    ax.margins(0.)
    ax.plot(rnks, es, color=color, linewidth=2)
    ax.axvline(rnks[j], linestyle='--', color=color)
    ax.axhline(0, linestyle='--', color=color)
    ax.set_ylabel('Enrichment\nScore')
    ax.set_title(name)
    # Plot gset mask
    ax = axes[1]
    ax.margins(0.)
    ax.set_yticklabels([])
    ax.set_yticks([])
    ax.vlines(rnks[set_msk], 0, 1, linewidth=0.5, color=color)
    # Plot color bar
    ax = axes[2]
    ax.margins(0.)
    ax.set_yticklabels([])
    ax.set_yticks([])
    vmin = np.percentile(np.min(m), 2)
    vmax = np.percentile(np.max(m), 98)
    norm = MidpointNormalize(vmin=vmin, vcenter=0, vmax=vmax)
    ax.pcolormesh(
        m[np.newaxis, :],
        rasterized=True,
        norm=norm,
        cmap=cmap,
    )
    ax.set_xlim(0, rnks.size-1)  # Remove extreme to the right
    # Plot ranks
    ax = axes[3]
    ax.margins(0.)
    ax.fill_between(rnks, y1=m, y2=0, color="#C9D3DB")
    non_zero_rnks = rnks[m > 0]
    if non_zero_rnks.size == 0:
        zero_rnk = rnks[-1]
    else:
        zero_rnk = non_zero_rnks[-1] + 1
    ax.axvline(zero_rnk, linestyle='--', color="#C9D3DB")
    ax.set_xlabel('Rank')
    ax.set_ylabel('Ranked\nmetric')
    # Remove spaces
    bp.fig.subplots_adjust(wspace=0, hspace=0)
    return bp._return(), le_c



import pandas as pd
import numpy as np
import scipy.sparse as sps
import pytest
from anndata import AnnData

import decoupler as dc


@pytest.mark.parametrize(
    'key',
    ['X_pca', 'X_umap', 'score_ulm', 'padj_ulm']
)
def test_get_obsm(
    tdata_obsm,
    key,
):
    obsm = dc.pp.get_obsm(adata=tdata_obsm, key=key)
    assert isinstance(obsm, AnnData)
    assert obsm.n_obs == tdata_obsm.n_obs
    assert obsm.n_vars == tdata_obsm.obsm[key].shape[1]
    assert (obsm.obs == tdata_obsm.obs).values.all()
    assert (obsm.X == tdata_obsm.obsm[key]).all().all()


def test_swap_layer(
    adata,
):
    ldata = adata.copy()
    res = dc.pp.swap_layer(adata=ldata, key='counts', X_key=None, inplace=True)
    assert res is None
    assert ldata.X.sum().is_integer()
    assert list(ldata.layers.keys()) == ['counts']
    ldata = adata.copy()
    res = dc.pp.swap_layer(adata=ldata, key='counts', X_key=None, inplace=False)
    assert isinstance(res, AnnData)
    assert res.X.sum().is_integer()
    assert list(res.layers.keys()) == ['counts']
    res = dc.pp.swap_layer(adata=ldata, key='counts', X_key='X', inplace=False)
    assert isinstance(res, AnnData)
    assert res.X.sum().is_integer()
    assert list(res.layers.keys()) == ['counts', 'X']
    assert (ldata.X == res.layers['X']).all()


@pytest.mark.parametrize(
    'groups_col,mode,sparse,empty',
    [
        [None, 'sum', False, True],
        [None, 'mean', True, True],
        ['group', 'median', False, False],
        ['group', lambda x: np.max(x) - np.min(x), True, True],
        ['group', dict(sum=np.sum, mean=np.mean), False, False],
    ]
)
def test_pseudobulk(
    adata,
    groups_col,
    mode,
    sparse,
    empty,
):
    adata = adata.copy()
    if empty:
        adata.X[:, 3] = 0.
        adata.layers['counts'][:, 3] = 0.
    if sparse:
        adata.X = sps.csr_matrix(adata.X)
    if mode == 'sum':
        layer = 'counts'
    else:
        layer = None
    pdata = dc.pp.pseudobulk(
        adata=adata,
        sample_col='sample',
        groups_col=groups_col,
        mode=mode,
        empty=empty,
        layer=layer,
        skip_checks=False,
    )
    assert isinstance(pdata, AnnData)
    assert pdata.shape[0] < adata.shape[0]
    if empty:
        assert pdata.shape[1] < adata.shape[1]
    else:
        assert pdata.shape[1] == adata.shape[1]
    obs_cols = {'psbulk_cells', 'psbulk_counts'}
    assert obs_cols.issubset(pdata.obs.columns)
    assert 'psbulk_props' in pdata.layers
    prop = pdata.layers['psbulk_props']
    assert ((0. <= prop) & (prop <= 1.)).all()
    if sparse:
        assert isinstance(pdata.X, np.ndarray)
    if isinstance(mode, dict):
        assert set(mode.keys()).issubset(pdata.layers.keys())


@pytest.mark.parametrize('inplace', [True, False])
def test_filter_samples(
    pdata,
    inplace,
):
    f_pdata = pdata.copy()
    res = dc.pp.filter_samples(adata=f_pdata, min_counts=90, min_cells=4, inplace=inplace)
    if inplace:
        assert res is None
        assert f_pdata.shape[0] < pdata.shape[0]
    else:
        assert isinstance(res, np.ndarray)
        assert res.size < pdata.shape[0]


def test_filter_by_expr(
    pdata,
):
    """
    names_v <- c(
      'G11', 'G04', 'G05', 'G03', 'G07', 'G18', 'G17','G02','G10', 'G14',
      'G09', 'G16', 'G08', 'G13', 'G20', 'G01', 'G12', 'G15', 'G06', 'G19'
    )
    names_o <- c(
      'S01_A', 'S02_A', 'S03_A', 'S01_B', 'S02_B', 'S03_B'
    )
    data <- c(
    0., 0., 2., 0., 2., 0., 2.,17., 0., 3., 0., 3., 0., 3., 3.,18., 0., 0., 1., 0.,
    0.,35., 3.,44., 2., 6., 7.,26., 3., 6., 0., 6., 1., 5., 5.,24., 1., 3., 4., 5.,
    2., 0., 0.,10., 1., 4., 6.,25., 1., 5., 0., 2., 0., 3., 8.,35., 2., 2., 0.,13.,
    2., 0., 9., 1., 9., 3., 3., 0., 1., 4., 0., 0.,19., 0., 0., 0., 0., 3., 8., 4.,
    0., 1., 8., 1.,19., 0., 7., 2., 0., 7., 1., 2.,24., 3.,10., 3., 0., 5.,17., 2.,
    2., 0.,34., 4.,42., 3., 3., 1., 3.,10., 1., 0.,28., 6., 9., 0., 3., 4.,17., 5.
    )
    data <- matrix(data = data, byrow = TRUE, nrow = length(names_o))
    rownames(data) <- names_o
    colnames(data) <- names_v
    data <- t(data)
    group <- c('A', 'A', 'A', 'B', 'B', 'B')
    msk <- filterByExpr(
      y=data, group = group, lib.size = NULL, min.count = 10,
      min.total.count = 10, large.n = 10, min.prop = 0.7)
    rownames(data)[msk]
    msk <- filterByExpr(
      y=data, group = group, lib.size = NULL, min.count = 7,
      min.total.count = 10, large.n = 10, min.prop = 0.7)
    rownames(data)[msk]
    msk <- filterByExpr(
      y=data, group = group, lib.size = NULL, min.count = 7,
      min.total.count = 10, large.n = 0, min.prop = 0.1)
    rownames(data)[msk]
    msk <- filterByExpr(
      y=data, group = group, lib.size = 1, min.count = 3,
      min.total.count = 10, large.n = 0, min.prop = 0.1)
    rownames(data)[msk]
    """
    dc_var = dc.pp.filter_by_expr(
        adata=pdata, group='group', lib_size=None, min_count=10, min_total_count=10, large_n=10, min_prop=0.7, inplace=False
    )
    eg_var = np.array(["G07", "G02", "G08", "G01", "G06"])
    assert set(dc_var) == set(eg_var)
    dc_var = dc.pp.filter_by_expr(
        adata=pdata, group='group', lib_size=None, min_count=7, min_total_count=10, large_n=10, min_prop=0.7, inplace=False
    )
    eg_var = np.array(["G05", "G07", "G02", "G08", "G01", "G06"])
    assert set(dc_var) == set(eg_var)
    dc_var = dc.pp.filter_by_expr(
        adata=pdata, group='group', lib_size=None, min_count=7, min_total_count=10, large_n=0, min_prop=0.1, inplace=False
    )
    eg_var = np.array(["G04", "G05", "G03", "G07", "G17", "G02", "G14", "G08", "G20", "G01", "G06", "G19"])
    assert set(dc_var) == set(eg_var)
    dc_var = dc.pp.filter_by_expr(
        adata=pdata, group='group', lib_size=1, min_count=3, min_total_count=10, large_n=0, min_prop=0.1, inplace=False
    )
    eg_var = np.array([
        "G04", "G05", "G03", "G07", "G18", "G17", "G02", "G14", "G16", "G08", "G13", "G20", "G01", "G15", "G06", "G19"
    ])
    assert set(dc_var) == set(eg_var)


@pytest.mark.parametrize('inplace', [True, False])
def test_filter_by_prop(
    pdata,
    inplace,
):
    f_pdata = pdata.copy()
    res = dc.pp.filter_by_prop(adata=f_pdata, inplace=inplace)
    if inplace:
        assert res is None
        assert f_pdata.shape[1] < pdata.shape[1]
    else:
        assert isinstance(res, np.ndarray)
        assert res.size < pdata.shape[1]


@pytest.mark.parametrize('key', ['X_pca', 'score_ulm'])
def test_knn(
    tdata_obsm,
    key,
):
    k_adata = tdata_obsm.copy()
    res = dc.pp.knn(adata=k_adata, key=key)
    assert res is None
    k = f'{key}_connectivities'
    assert k in k_adata.obsp
    assert isinstance(k_adata.obsp[k], sps.csr_matrix)


@pytest.mark.parametrize(
    'names,order,label,nbins',
    [
        [None, 'pstime', None, 10],
        ['G05', 'pstime', None, 35],
        [['G05'], 'f_order', None, 10],
        [['G01', 'G06', 'G10'], 'pstime', 'group', 14],
    ]
)
def test_bin_order(
    tdata,
    names,
    order,
    label,
    nbins,
):
    rng = np.random.default_rng(seed=42)
    tdata.obs.loc[:, 'f_order'] = rng.random(tdata.n_obs)
    df = dc.pp.bin_order(adata=tdata, names=names, order=order, label=label, nbins=nbins)
    assert isinstance(df, pd.DataFrame)
    cols = {'name', 'order', 'value'}
    assert cols.issubset(df.columns)
    assert ((0. <= df['order']) & (df['order'] <= 1.)).all()
    assert df['order'].unique().size == np.min([tdata.n_obs, nbins])
    if label is not None:
        lcols = {'label', 'color'}
        assert lcols.issubset(df.columns)
        s_lbl = set(tdata.obs[label])
        assert s_lbl  == set(df['label'])
        assert len(s_lbl) == len(set(df['color']))
        assert f'{label}_colors' in tdata.uns.keys()



import pandas as pd
import pytest

import decoupler as dc


@pytest.mark.parametrize(
    'key,uns_key',
    [
        ['X_pca', 'rank_obsm'],
        ['X_pca', None],
        ['X_umap', 'other'],
        ['score_ulm', 'other'],
        ['score_ulm', None],
    ]
)
def test_rankby_obsm(
    tdata_obsm,
    key,
    uns_key,
):
    res = dc.tl.rankby_obsm(tdata_obsm, key=key, uns_key=uns_key)
    if uns_key is None:
        assert isinstance(res, pd.DataFrame)
    else:
        assert res is None
        assert uns_key in tdata_obsm.uns
        assert isinstance(tdata_obsm.uns[uns_key], pd.DataFrame)



import numpy as np
import pytest

import decoupler as dc


@pytest.mark.parametrize(
    'fun,times,seed',
    [
        ['wmean', 10, 42],
        ['wsum', 5, 23],
        [lambda x, w: 0, 5, 1],
        ['wmean', 0, 42],
    ]
)
def test_func_waggr(
    mat,
    adjmat,
    fun,
    times,
    seed,
):
    X, obs, var = mat
    es, pv = dc.mt._waggr._func_waggr(mat=X, adj=adjmat, fun=fun, times=times, seed=seed)
    assert np.isfinite(es).all()
    assert ((0 <= pv) & (pv <= 1)).all()


import pandas as pd
import scipy.sparse as sps
import pytest

import decoupler as dc


def test_return(
    adata,
    net,
):
    mth = dc.mt.ulm
    adata = adata[:4].copy()
    adata.X[:, 0] = 0.
    es, pv = mth(data=adata.to_df(), net=net, tmin=0)
    r = dc.mt._run._return(name=mth.name, data=adata, es=es, pv=pv)
    assert r is None
    r = dc.mt._run._return(name=mth.name, data=adata.to_df(), es=es, pv=pv)
    assert isinstance(r, tuple)
    assert isinstance(r[0], pd.DataFrame)
    assert isinstance(r[1], pd.DataFrame)


@pytest.mark.parametrize(
    'mth,bsize',
    [
        [dc.mt.zscore, 2],
        [dc.mt.ora, 2],
        [dc.mt.gsva, 250_000],
    ]
)
def test_run(
    adata,
    net,
    mth,
    bsize,
):
    sdata = adata.copy()
    sdata.X = sps.csr_matrix(sdata.X)
    des, dpv = dc.mt._run._run(
        name=mth.name,
        func=mth.func,
        adj=mth.adj,
        test=mth.test,
        data=adata.to_df(),
        net=net,
        tmin=0,
    )
    ses, spv = dc.mt._run._run(
        name=mth.name,
        func=mth.func,
        adj=mth.adj,
        test=mth.test,
        data=sdata.to_df(),
        net=net,
        tmin=0,
    )
    assert (des.values == ses.values).all()
    


from typing import Tuple

import pandas as pd
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from anndata import AnnData
import marsilea as ma
import marsilea.plotter as mp

from decoupler._docs import docs
from decoupler._Plotter import Plotter


def _input(
    adata: AnnData,
    uns_key: str,
    names: str | list | None = None,
    nvar: int | float | list | None = 10,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    assert isinstance(adata, AnnData), 'adata must be adata.AnnData'
    assert isinstance(uns_key, str) and uns_key in adata.uns, \
    'uns_key must be str and in adata.uns'
    assert isinstance(names, (str, list)) or names is None, \
    'names must be str, list or None'
    assert isinstance(nvar, (int, float, str, list)) or nvar is None, \
    'nvar must be numeric, list or None'
    # Filter stats by obs names
    stats = adata.uns[uns_key]
    assert hasattr(stats, 'key'), \
    'adata.uns[key] must be generated by decoupler.tl.rankby_obsm'
    obsm_key = stats.key
    stats = stats.sort_values('obsm')
    if isinstance(names, str):
        names = [names]
    if isinstance(nvar, str):
        nvar = [nvar]
    if names:
        stats = stats[stats['obs'].isin(names)]
    # Filter stats by obsm nvar
    obsm = adata.obsm[obsm_key]
    if isinstance(obsm, pd.DataFrame):
        var_names = obsm.std(ddof=1, axis=0).sort_values(ascending=False).index
        if isinstance(nvar, (int, float)):
            var_names = var_names[:int(nvar)]
        elif isinstance(nvar, list):
            var_names = nvar
        obsm = obsm.loc[:, var_names]
        stats = stats[stats['obsm'].isin(var_names)]
    else:
        var_names = sorted(stats['obsm'].unique())
        if isinstance(nvar, (int, float)):
            var_names = var_names[:nvar]
            obsm = obsm[:, :nvar]
        elif isinstance(nvar, list):
            idx = np.searchsorted(var_names, nvar)
            obsm = obsm[:, idx]
            var_names = nvar
        stats = stats[stats['obsm'].isin(var_names)]
    # Extract obsm
    obsm = pd.DataFrame(obsm, columns=var_names)
    # Transform stats
    min_p = stats[stats['padj'] > 0]['padj'].min()
    stats.loc[stats['padj'] == 0, 'padj'] = min_p
    stats['padj'] = -np.log10(stats['padj'])
    stats = stats.pivot(index='obs', columns='obsm', values='padj')
    stats.index.name = None
    stats.columns.name = None
    if names is None:
        names = stats.index
    return obsm, stats, names


@docs.dedent
def obsm(
    adata: AnnData,
    key: str = 'rank_obsm',
    names: str | list | None = None,
    nvar: int | float | str | list | None = 10,
    dendrogram: bool = True,
    thr_sign: float = 0.05,
    titles: list = ['Scores', 'Stats'],
    cmap_stat: str = 'Purples',
    cmap_obsm: str = 'BrBG',
    cmap_obs: dict = dict(),
    **kwargs
) -> None | Figure:
    """
    Plot metadata associations with features in ``adata.obsm``.

    Parameters
    ----------
    %(adata)s
    key
        Name of ``adata.uns`` key storing ``decoupler.tl.rank_obsm`` results.
    names
        Which metadata covariates to show.
    nvar
        How many features from ``adata.obsm`` to show.
    dendogram
        Whether to sort and plot samples using a dendogram.
    thr_sign
        Threshold of significance for the adjusted p-values.
    titles
        List of titles to place for the metadata heatmap and ``obsm`` features.
    cmap_stat
        Colormap for metadata statistics.
    cmap_obsm
        Colormap for ``obsm`` features.
    cmap_obs
        Dictionary of colormaps containing a palette for each metadata covariate being plotted.
    %(plot)s
    """
    assert isinstance(dendrogram, bool), 'dendrogram must be bool'
    assert isinstance(thr_sign, float) and 1 >= thr_sign >= 0, \
    'thr_sign must be float and between 0 and 1'
    assert isinstance(titles, list) and len(titles) == 2, \
    'titles must be list and with 2 elements'
    assert isinstance(cmap_obs, dict), \
    'cmap_obs must be dict'
    # Extract
    obsm, stats, names = _input(adata=adata, uns_key=key, names=names, nvar=nvar)
    # Instance
    kwargs['ax'] = None
    bp = Plotter(**kwargs)
    bp.fig.delaxes(bp.ax)
    plt.close(bp.fig)
    # Plot stats
    h1 = ma.Heatmap(stats, cmap=cmap_stat, name="h1", width=4, height=1, label=r'$-\log_{10}(padj)$')
    h1.add_title(top=titles[1], align="center")
    if dendrogram:
        h1.add_dendrogram("left")
    h1.add_right(mp.Labels(stats.index, ))
    sign_msk = stats.values > -np.log10(thr_sign)
    layer = mp.MarkerMesh(sign_msk, marker='*', label=f"padj < {thr_sign}", color='red')
    h1.add_layer(layer, name='sign')
    # Plot obsm
    h2 = ma.Heatmap(obsm, cmap=cmap_obsm, name="h2", width=0.4, height=4, label='Score')
    h2.add_title(top=titles[0], align="center")
    if dendrogram:
        h2.add_dendrogram("left")
    h2.add_bottom(mp.Labels(obsm.columns))
    # Add obs legends
    for name in names:
        is_numeric = pd.api.types.is_numeric_dtype(adata.obs[name])
        if is_numeric:
            if name not in cmap_obs:
                cmap = 'viridis'
            else:
                cmap = cmap_obs[name]
            colors = mp.ColorMesh(adata.obs[name], cmap=cmap, label=name)
        else:
            cats = adata.obs[name].sort_values().unique()
            if name not in cmap_obs:
                tab10 = plt.get_cmap('tab10')
                palette = {k: tab10(i) for i, k in enumerate(cats)}
            else:
                c_cmap = plt.get_cmap(cmap_obs[name])
                palette = {k: c_cmap(i) for i, k in enumerate(cats)}
            colors = mp.Colors(adata.obs[name], palette=palette, label=name)
        h2.add_right(colors, pad=0.1, size=0.1)
    # Build plot
    c = (h1 / .05 / h2)
    c.add_legends(side='right', stack_by='row', stack_size=3, align_legends='top')
    c.render()
    if bp.return_fig or bp.save is not None:
        plt.close()
    # Add borders
    hax = c.get_ax(board_name='h1', ax_name='h1')
    border = matplotlib.patches.Rectangle((0, 0), 1, 1, fill=False, ec=".1", lw=2, transform=hax.transAxes)
    hax.add_artist(border)
    hax = c.get_ax(board_name='h2', ax_name='h2')
    border = matplotlib.patches.Rectangle((0, 0), 1, 1, fill=False, ec=".1", lw=2, transform=hax.transAxes)
    hax.add_artist(border)
    bp.fig = c.figure
    bp.fig.set_figwidth(bp.figsize[0])
    bp.fig.set_figheight(bp.figsize[1])
    bp.fig.set_dpi(bp.dpi)
    return bp._return()



from typing import Callable
import textwrap

import pandas as pd

from decoupler._docs import docs
from decoupler._datatype import DataType
from decoupler.mt._run import _run


class MethodMeta:
    def __init__(
        self,
        name: str,
        desc: str,
        func: Callable,
        stype: str,
        adj: bool,
        weight: bool,
        test: bool,
        limits: tuple,
        reference: str,
    ):
        self.name = name
        self.desc = desc
        self.func = func
        self.stype = stype
        self.adj = adj
        self.weight = weight
        self.test = test
        self.limits = limits
        self.reference = reference

    def meta(self) -> pd.DataFrame:
        meta = pd.DataFrame([{
            'name': self.name,
            'desc': self.desc,
            'stype': self.stype,
            'weight': self.weight,
            'test': self.test,
            'limits': self.limits,
            'reference': self.reference
        }])
        return meta


#@docs.dedent
class Method(MethodMeta):
    def __init__(
        self,
        _method: MethodMeta,
    ):
        super().__init__(
            name=_method.name,
            desc=_method.desc,
            func=_method.func,
            stype=_method.stype,
            adj=_method.adj,
            weight=_method.weight,
            test=_method.test,
            limits=_method.limits,
            reference=_method.reference,
        )
        self._method = _method
        self.__doc__ = self.func.__doc__

    def __call__(
        self,
        data: DataType,
        net: pd.DataFrame,
        tmin: int | float = 5,
        raw: bool = False,
        empty: bool = True,
        bsize: int | float = 250_000,
        verbose: bool = False,
        **kwargs,
    ):
        return _run(
            name=self.name,
            func=self.func,
            adj=self.adj,
            test=self.test,
            data=data,
            net=net,
            tmin=tmin,
            raw=raw,
            empty=empty,
            bsize=bsize,
            verbose=verbose,
            **kwargs,
        )


def _show_methods(methods):
    return pd.concat([method.meta() for method in methods]).reset_index(drop=True)



import pandas as pd
import numpy as np
import pytest

import decoupler as dc


@pytest.mark.parametrize(
    'top,thr_padj',
    [
        [100, 0.05],
        [100, 1],
        [np.inf, 0.05],
        [np.inf, 1],
    ]
)
def test_progeny(
    top,
    thr_padj,
):
    pr = dc.op.progeny(top=top, thr_padj=thr_padj)
    assert isinstance(pr, pd.DataFrame)
    cols = {'source', 'target', 'weight', 'padj'}
    assert cols.issubset(pr.columns)
    assert pd.api.types.is_numeric_dtype(pr['weight'])
    assert pd.api.types.is_numeric_dtype(pr['padj'])
    assert (pr['padj'] < thr_padj).all()
    assert (pr.groupby('source').size() <= top).all()
    assert not pr.duplicated(['source', 'target']).any()



from typing import Tuple

import pandas as pd
import numpy as np
from numpy.random import default_rng
from anndata import AnnData
import scipy.sparse as sps

from decoupler._docs import docs
from decoupler._log import _log
from decoupler._datatype import DataType


def _extract(
    data: DataType,
    layer: str | None = None,
    raw: bool = False,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    assert isinstance(data, (list, pd.DataFrame, AnnData)), \
    'mat must be a list of [matrix, samples, features], pd.DataFrame (samples x features)\n\
    or an AnnData instance'
    assert layer is None or isinstance(layer, str), 'layer must be str or None'
    assert isinstance(raw, bool), 'raw must be bool'
    if isinstance(data, list):
        mat, row, col = data
        mat = np.array(mat)
        row = np.array(row, dtype='U')
        col = np.array(col, dtype='U')
    elif isinstance(data, pd.DataFrame):
        mat = data.values.astype(float)
        row = data.index.values.astype('U')
        col = data.columns.values.astype('U')
    elif isinstance(data, AnnData):
        row = data.obs_names.values.astype('U')
        if raw:
            assert data.raw, 'Received `raw=True`, but `mat.raw` is empty.'
            mat = data.raw.X.astype(float)
            col = data.raw.var_names.values.astype('U')
        else:
            col = data.var_names.values.astype('U')
            if layer:
                mat = data.layers[layer].astype(float)
            else:
                mat = data.X.astype(float)
    return mat, row, col


def _validate_mat(
    mat: np.ndarray,
    row: np.ndarray,
    col: np.ndarray,
    empty: bool = True,
    verbose: bool = False
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    assert isinstance(empty, bool), 'empty must be bool'
    # Accept any sparse format but transform to csr
    if sps.issparse(mat) and not isinstance(mat, sps.csr_matrix):
        mat = sps.csr_matrix(mat)
    # Check for empty features
    if isinstance(mat, sps.csr_matrix):
        msk_col = mat.getnnz(axis=0) == 0
    else:
        msk_col = np.count_nonzero(mat, axis=0) == 0
    n_empty_col = np.sum(msk_col)
    if n_empty_col > 0 and empty:
        m = f'{n_empty_col} features of mat are empty, they will be removed'
        _log(m, level='warn', verbose=verbose)
        col = col[~msk_col]
        mat = mat[:, ~msk_col]
    # Check for repeated features
    assert not np.any(col[1:] == col[:-1]), \
    'mat contains repeated feature names, please make them unique'
    # Check for empty samples
    if isinstance(mat, sps.csr_matrix):
        msk_row = mat.getnnz(axis=1) == 0
    else:
        msk_row = np.count_nonzero(mat, axis=1) == 0
    n_empty_row = np.sum(msk_row)
    if n_empty_row > 0 and empty:
        m = f'{n_empty_row} observations of mat are empty, they will be removed'
        _log(m, level='warn', verbose=verbose)
        row = row[~msk_row]
        mat = mat[~msk_row]
    # Check for non finite values
    assert not np.any(~np.isfinite(mat.data)), \
    'mat contains non finite values (nan or inf), set them to 0 or remove them'
    return mat, row, col


def break_ties(
    mat: np.ndarray,
    features: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    # Randomize feature order to break ties randomly
    rng = default_rng(seed=0)
    idx = np.arange(features.size)
    idx = rng.choice(idx, features.size, replace=False)
    mat, features = mat[:, idx], features[idx]
    return mat, features


@docs.dedent
def extract(
    data: DataType,
    layer: str | None = None,
    raw: bool = False,
    empty: bool = True,
    verbose: bool = False,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Extracts matrix, rownames and colnames from data.

    Parameters
    ----------
    %(data)s
    %(layer)s
    %(raw)s
    %(empty)s
    %(verbose)s

    Returns
    -------
    Matrix, rownames and colnames from data.
    """
    # Extract
    mat, row, col = _extract(data=data, layer=layer, raw=raw)
    m = f'Extracted omics mat with {row.size} rows (observations) and {col.size} columns (features)'
    _log(m, level='info', verbose=verbose)
    # Validate
    mat, row, col = _validate_mat(mat=mat, row=row, col=col, empty=empty, verbose=verbose)
    # Randomly sort features
    mat, col = break_ties(mat=mat, features=col)
    return mat, row, col



import pandas as pd
import seaborn as sns
from matplotlib.figure import Figure

from decoupler._docs import docs
from decoupler._Plotter import Plotter
from decoupler.bm.pl._format import _format


@docs.dedent
def fscore(
    df: pd.DataFrame,
    hue: str | None = None,
    palette: str = 'tab20',
    **kwargs
) -> None | Figure:
    """
    Plot precision and recall as scatterplot.

    x-axis represent the recall of correctly predicted sources after filtering by significance.
    The higher value the better performance is.

    x-axis represent the precision of correctly predicted sources after filtering by significance.
    The higher value the better performance is.

    Parameters
    ----------
    %(df)s
    %(hue)s
    %(palette)s
    %(plot)s
    """
    # Validate
    assert isinstance(hue, str) or hue is None, 'hue must be str or None'
    # Format
    tmp = _format(df=df, cols=['recall', 'precision'])
    # Instance
    bp = Plotter(**kwargs)
    # Plot
    args = dict()
    if hue is not None:
        args['hue'] = hue
        args['palette'] = palette
    sns.scatterplot(
        data=tmp,
        x='recall',
        y='precision',
        ax=bp.ax,
        **args
    )
    if hue is not None:
        bp.ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False, title=hue)
    return bp._return()



from typing import Tuple

import pandas as pd
import numpy as np
import scipy.stats as sts
from tqdm.auto import tqdm
import numba as nb

from decoupler._docs import docs
from decoupler._log import _log
from decoupler._datatype import DataType
from decoupler.pp.data import extract


def read_gmt(
    path: str,
) -> pd.DataFrame:
    """
    Read a GMT file and return the feature sets as a network.

    Parameters
    ----------
    path
        Path to GMT file containing feature sets.

    Returns
    -------
    Gene sets as ``pd.DataFrame``.
    """
    # Init empty df
    df = []
    # Read line per line
    with open(path, 'r') as f:
        for line in f.readlines():
            line = line.rstrip().split()
            # Extract gene set name
            set_name = line[0]
            # For each gene add an entry (skip link in [1])
            genes = line[2:]
            for gene in genes:
                df.append([set_name, gene])
    # Transform to df
    df = pd.DataFrame(df, columns=['source', 'target'])
    return df


def _validate_net(
    net = pd.DataFrame,
    verbose: bool = False,
) -> pd.DataFrame:
    assert isinstance(net, pd.DataFrame), 'net must be a DataFrame'
    assert {'source', 'target'}.issubset(net.columns), \
    "DataFrame must have 'source' and 'target' columns\n \
    If present but with a different names use:\n \
    net = net.rename(columns={'...' : 'source', '...': 'target'})"
    assert not net.duplicated(subset=['source', 'target']).any(), \
    "net has duplicate rows, use:\n \
    net = net.drop_duplicates(subset=['source', 'target'])"
    if 'weight' not in net.columns:
        vnet = net[['source', 'target']].copy()
        vnet['weight'] = 1.
        m = "weight not found in net.columns, adding it as:\nnet['weight'] = 1"
        _log(m, level='warn', verbose=verbose)
    else:
        vnet = net[['source', 'target', 'weight']].copy()
    vnet['source'] = vnet['source'].astype('U')
    vnet['target'] = vnet['target'].astype('U')
    vnet['weight'] = vnet['weight'].astype(float)
    return vnet


@docs.dedent
def prune(
    features: np.ndarray,
    net: pd.DataFrame,
    tmin: int = 5,
    verbose: bool = False,
) -> pd.DataFrame:
    """
    Removes sources of a ``net`` with less than ``tmin`` targets shared with ``mat``.

    Parameters
    ----------
    %(features)s
    %(net)s
    %(tmin)s
    %(verbose)s

    Returns
    -------
    Filtered net in long format.
    """
    # Validate
    vnet = _validate_net(net, verbose=verbose)
    features = set(features)
    assert isinstance(tmin, (int, float)) and tmin >= 0, 'tmin must be numeric and >= 0'
    # Find shared targets between mat and net
    msk = vnet['target'].isin(features)
    vnet = vnet.loc[msk]
    # Find unique sources with tmin
    sources = vnet['source'].value_counts()
    sources = set(sources[sources >= tmin].index)
    # Filter
    msk = vnet['source'].isin(sources)
    vnet = vnet[msk]
    assert not vnet.empty, \
    f'No sources with more than tmin={tmin} targets after\n \
    filtering by shared features in mat.\n \
    Make sure mat and net have shared target features or\n \
    reduce the number assigned to tmin'
    return vnet


def _adj(
    net: pd.DataFrame,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    # Pivot df to a wider format
    X = net.pivot(columns='source', index='target', values='weight').fillna(0)
    # Store node names and weights
    sources = X.columns.values.astype('U')
    targets = X.index.values.astype('U')
    X = X.values.astype(float)
    return sources, targets, X


def _order(
    features: np.ndarray,
    targets: np.ndarray,
    adjmat: np.ndarray,
) -> np.ndarray:
    # Init empty madjmat
    madjmat = np.zeros((len(features), adjmat.shape[1]))
    # Create an index array for rows of features corresponding to targets
    features_dict = {gene: i for i, gene in enumerate(features)}
    idxs = [features_dict[gene] for gene in targets if gene in features_dict]
    assert len(idxs) > 0, 'No overlap found between features and targets'
    # Populate madjmat using advanced indexing
    madjmat[idxs, :] = adjmat[: len(idxs), :]
    return madjmat


@docs.dedent
def adjmat(
    features: np.ndarray,
    net: pd.DataFrame,
    verbose: bool = False,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Converts a network in long format into a regulatory adjacency matrix (targets x sources).

    Parameters
    ----------
    %(net)s

    Returns
    -------
    Returns the source names (columns), target names (rows), and the adjacency matrix of weights.
    """
    # Extract adj mat
    sources, targets, adjm = _adj(net=net)
    # Sort adjmat to match features
    adjm = _order(features, targets, adjm)
    m = f'Network adjacency matrix has {targets.size} unique features and {sources.size} unique sources'
    _log(m, level='info', verbose=verbose)
    return sources, targets, adjm


@docs.dedent
def idxmat(
    features: np.ndarray,
    net: pd.DataFrame,
    verbose: bool = False,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Indexes and returns feature sets as a decomposed sparse matrix.

    Parameters
    ----------
    %(features)s
    %(net)s

    Returns
    -------
    List of sources, concatenated indexes, starts and offsets.
    """
    # Transform targets to indxs
    table = {name: i for i, name in enumerate(features)}
    net['idx_target'] = [table[target] for target in net['target']]
    # Find sets
    cnct = (
        net
        .groupby('source', observed=True)
        ['idx_target']
        .apply(lambda x: np.array(x, dtype=int))
    )
    net.drop(columns=['idx_target'], inplace=True)
    sources = cnct.index.values.astype('U')
    # Flatten net and get offsets
    offsets = cnct.apply(lambda x: len(x)).values
    cnct = np.concatenate(cnct.values)
    # Define starts to subset offsets
    starts = np.zeros(offsets.shape[0], dtype=int)
    starts[1:] = np.cumsum(offsets)[:-1]
    targets = np.unique(cnct)
    m = f'Network has {targets.size} unique features and {sources.size} unique sources'
    _log(m, level='info', verbose=verbose)
    return sources, cnct, starts, offsets


@nb.njit(cache=True)
def _getset(
    cnct: np.ndarray,
    starts: np.ndarray,
    offsets: np.ndarray,
    j: int
) -> np.ndarray:
    srt = starts[j]
    off = srt + offsets[j]
    fset = cnct[srt:off]
    return fset


@docs.dedent
def shuffle_net(
    net: pd.DataFrame,
    target: bool = True,
    weight: bool = False,
    seed: int = 42,
    same_seed: bool = True
) -> pd.DataFrame:
    """
    Shuffle a network to make it random.

    Shuffle a given net by targets, weight or both at the same time.

    If only targets are shuffled, targets will change but the
    distribution of weights for each set will be preserved.

    If only weights are shuffled, targets will be the same but the
    distribution of weights for each set will change.

    If targets and weights are shuffled at the same time,
    both targets and weight distribution will change for each set.

    Parameters
    ----------
    %(net)s
    target
        Whether to shuffle targets.
    weight
        Whether to shuffle weights.
    %(seed)s
    same_seed : bool
        Whether to share seed when shuffling targets and weights.

    Returns
    -------
    Shuffled network.
    """
    # Validate
    assert isinstance(net, pd.DataFrame), 'net must be pandas.DataFrame'
    assert isinstance(target, bool), 'target must be bool'
    assert isinstance(weight, bool), 'weight must be bool'
    assert target or weight, 'If target and weight are both False, nothing is shuffled'
    assert isinstance(same_seed, bool), 'same_seed must be bool'
    assert 'target' in net.columns, 'target must be in net.columns'
    assert 'weight' in net.columns, 'weight must be in net.columns'
    # Make copy of net
    rnet = net.copy()
    # Shuffle
    if target:
        rng = np.random.default_rng(seed=seed)
        rng.shuffle(rnet['target'].values)
    if weight:
        rng = np.random.default_rng(seed=seed + int(not same_seed))
        rng.shuffle(rnet['weight'].values)
    return rnet.drop_duplicates(['source', 'target'], keep='first')


@docs.dedent
def net_corr(
    net: pd.DataFrame,
    data: None | DataType = None,
    tmin: int = 5,
    verbose: bool = False,
    **kwargs
) -> pd.DataFrame:
    """
    Checks the correlation across the sources in a network.
    If data is also provided, target features will be prunned to
    match the ones in mat.

    Parameters
    ----------
    %(net)s
    %(data)s
    %(tmin)s
    %(verbose)s
    kwargs
        All other keyword arguments are passed to ``decoupler.pp.extract``

    Returns
    -------
    Correlation pairs dataframe.
    """
    net = _validate_net(net, verbose=verbose)
    # If mat is provided
    if data is not None:
        # Extract sparse matrix and array of genes
        kwargs.setdefault('verbose', verbose)
        _, _, c = extract(data=data, **kwargs)
    else:
        c = np.unique(net['target'].values).astype('U')
    net = prune(features=c, net=net, tmin=tmin)
    sources, targets, adj = adjmat(features=c, net=net, verbose=False)
    # Compute corr
    corr = []
    for i, s_a in enumerate(tqdm(sources[:-1], disable=not verbose)):
        idx = np.arange(i + 1)
        A = np.delete(adj, idx, axis=1)
        b = adj[:, i].reshape(-1, 1)
        r, p = sts.pearsonr(A, b)
        for j, s_b in enumerate(sources[i + 1:]):
            corr.append([s_a, s_b, r[j], p[j]])
    corr = pd.DataFrame(corr, columns=['source_a', 'source_b', 'corr', 'pval'])
    corr['padj'] = sts.false_discovery_control(corr['pval'])
    corr['abs_corr'] = corr['corr'].abs()
    corr = (
        corr
        .sort_values(['padj', 'pval', 'abs_corr'], ascending=[False, True, True])
        .reset_index(drop=True)
        .drop(columns=['abs_corr'])
    )
    return corr



from typing import Tuple

import pandas as pd
import numpy as np
from anndata import AnnData

from decoupler._log import _log
from decoupler.pp.net import prune
from decoupler.mt._methods import _methods
from decoupler.mt._decouple import decouple
from decoupler.bm._pp import _validate_groupby, _validate_obs, _filter, _sign
from decoupler.bm.metric import dict_metric


def _testsign(
    adata: AnnData,
    mth: str,
    test: bool,
    thr: float,
) -> np.ndarray:
    assert isinstance(thr, (int, float)) and 0. <= thr <= 1., \
    'thr must be numeric and between 0 and 1'
    if test:
        sign = adata.obsm[f'padj_{mth}'].values <= thr
        sign = (adata.obsm[f'score_{mth}'].values > 0) & sign
    else:
        q = np.quantile(adata.obsm[f'score_{mth}'].values, 1 - thr, axis=1).reshape(-1, 1)
        sign = adata.obsm[f'score_{mth}'].values > q
    return sign


def _tensor_scores(
    adata: AnnData,
    thr: float,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, list]:
    # Get unique methods
    has_test = {m.name: m.test for m in _methods}
    has_test = has_test | {'consensus': True}
    mthds = [k.replace('score_', '') for k in adata.obsm if k.startswith('score_')]
    # Extract dimensions
    exps = adata.obs_names
    srcs = adata.obsm[f'score_{mthds[0]}'].columns.values
    # Build acts tensor and sort by exps and srcs
    n_exp, n_src, n_mth = len(exps), len(srcs), len(mthds)
    scores = np.zeros((n_exp, n_src, n_mth))
    signs = np.zeros(scores.shape, dtype=np.bool_)
    for i, mth in enumerate(mthds):
        scores[:, :, i] = adata.obsm[f'score_{mth}'].values
        signs[:, :, i] = _testsign(adata=adata, mth=mth, test=has_test[mth], thr=thr)
    return scores, signs, srcs, mthds


def _tensor_truth(
    obs: pd.DataFrame,
    srcs: np.ndarray
) -> pd.DataFrame:
    # Explode nested perturbs and pivot into mat
    grts = obs.explode('source').pivot(columns='source', values='type_p').notna().astype(float).fillna(0.)
    miss_srcs = srcs[~np.isin(srcs, grts.columns)]
    miss_srcs = pd.DataFrame(0, index=grts.index, columns=miss_srcs)
    grts = pd.concat([grts, miss_srcs], axis=1)
    grts = grts.loc[:, srcs].values
    return grts


def _unique_obs(
    col: np.ndarray,
) -> np.ndarray:
    # Gets unique categories from a column with both lists and elements.
    cats = set()
    for row in col:
        # Check if col elements are lists
        if type(row) is list:
            for r in row:
                if r not in cats:
                    cats.add(r)
        else:
            if row not in cats:
                cats.add(row)
    return np.sort(list(cats))


def _mask_grps(
    obs: pd.DataFrame,
    groupby: None | list,
    verbose: float,
) -> Tuple[list, list, list]:
    if groupby is not None:
        # Init empty lsts
        msks = []
        grps = []
        grpbys = []
        for grpby_i in groupby:
            # Handle nested groupbys
            if type(grpby_i) is list:
                grpby_i = np.sort(grpby_i)
                grpby_name = '|'.join(grpby_i)
                if grpby_i.size > 1:
                    obs[grpby_name] = obs[grpby_i[0]].str.cat(obs[grpby_i[1:]], sep='|')
                grpby_i = grpby_name
            m = f'benchmark - grouping by {grpby_i}'
            _log(m, level='info', verbose=verbose)
            # Find msk in obs based on groupby
            grps_j = _unique_obs(obs[grpby_i].values)
            msk_i = []
            grps_i = []
            for grp in grps_j:
                m = np.array([grp in lst for lst in obs[grpby_i]])
                msk_i.append(m)
                grps_i.append(grp)
            # Append
            msks.append(msk_i)
            grpbys.append(grpby_i)
            grps.append(grps_i)
    else:
        m = f'benchmark - running without grouping'
        _log(m, level='info', verbose=verbose)
        msks = None
        grpbys = None
        grps = None
    return msks, grpbys, grps


def _metric(
    y_score: np.ndarray,
    y_true: np.ndarray,
    y_sign: np.ndarray,
    metric: str,
    verbose: bool,
    **kwargs
) -> dict:
    f = dict_metric[metric]
    if metric == 'fscore':
        y_score = y_sign
    score = f(y_true=y_true, y_score=y_score, **kwargs)
    return {k: s for k, s in zip(f.scores, score)}


def _metric_scores(
    df: pd.DataFrame,
    grpby_i: str | None,
    grp: str | None,
    score: np.ndarray,
    sign: np.ndarray,
    grt: np.ndarray,
    srcs: np.ndarray,
    mthds: list,
    metrics: list,
    runby: str,
    emin: int,
    verbose: bool,
    **kwargs
) -> None:
    if runby == 'expr':
        m = ('benchmark - evaluating by experiment on:\n' +
        f'n_expr={score.shape[0]}, n_sources={score.shape[1]} across metrics={metrics}')
        _log(m, level='info', verbose=verbose)
        for m in range(len(mthds)):
            mth = mthds[m]
            scr_i = score[:, :, m]
            sgn_i = sign[:, :, m]
            for metric in metrics:
                vals = _metric(
                    y_score=scr_i,
                    y_true=grt,
                    y_sign=sgn_i,
                    metric=metric,
                    verbose=verbose,
                    **kwargs
                )
                for cname, val in vals.items():
                    row = [grpby_i, grp, None, mth, cname, val]
                    df.append(row)
    elif runby == 'source':
        m = f'benchmark - evaluating by source'
        _log(m, level='info', verbose=verbose)
        for m in range(len(mthds)):
            mth = mthds[m]
            # Remove sources with less than emin
            src_msk = np.sum(grt > 0., axis=0) >= emin
            scr_i, sgn_i, grt_i = score[:, src_msk, :], sign[:, src_msk, :], grt[:, src_msk]
            srcs_method = srcs[src_msk]
            for s in range(len(srcs_method)):
                src = srcs_method[s]
                scr_source = scr_i[:, s, m]
                sgn_source = sgn_i[:, s, m]
                grt_source = grt_i[:, s]
                # Check that grt is not all the same
                unq_grt = np.unique(grt_source[~np.isnan(scr_source)])
                # Convert from vector to arr
                scr_source, sgn_source, grt_source = scr_source[np.newaxis], sgn_source[np.newaxis], grt_source[np.newaxis]
                if unq_grt.size > 1:
                    for metric in metrics:
                        vals = _metric(
                            y_score=scr_source,
                            y_true=grt_source,
                            y_sign=sgn_source,
                            metric=metric,
                            verbose=verbose,
                            **kwargs
                        )
                        for cname, val in vals.items():
                            row = [grpby_i, grp, src, mth, cname, val]
                            df.append(row)


def _eval_scores(
    adata: pd.DataFrame,
    groupby: None | list,
    runby: str,
    metrics: str | list,
    thr: float,
    emin: int,
    verbose: bool,
    **kwargs
) -> pd.DataFrame:
    scores, signs, srcs, mthds = _tensor_scores(adata=adata, thr=thr)
    grts = _tensor_truth(obs=adata.obs, srcs=srcs)
    msks, grpbys, grps = _mask_grps(obs=adata.obs, groupby=groupby, verbose=verbose)
    df = []
    if msks is not None:
        n_grpbys = len(msks)
        for i in range(n_grpbys):
            msk_i = msks[i]
            grpby_i = grpbys[i]
            grps_i = grps[i]
            n_grps = len(grps_i)
            for j in range(n_grps):
                msk = msk_i[j]
                grp = grps_i[j]
                n = np.sum(msk)
                if n >= emin:
                    score, sign, grt = scores[msk, :, :], signs[msk, :, :], grts[msk, :]
                    # Special case when groupby == perturb, remove extra grts
                    if grp in srcs:
                        m = grp == srcs
                        grt[:, ~m] = 0.
                    # Compute and append scores to df
                    _metric_scores(
                        df=df,
                        grpby_i=grpby_i,
                        grp=grp,
                        score=score,
                        sign=sign,
                        grt=grt,
                        srcs=srcs,
                        mthds=mthds,
                        metrics=metrics,
                        runby=runby,
                        emin=emin,
                        verbose=verbose,
                        **kwargs
                    )
    else:
        n_exp = scores.shape[0]
        if n_exp >= emin:
            # Compute and append scores to df
            _metric_scores(
                df=df,
                grpby_i=None,
                grp=None,
                score=scores,
                sign=signs,
                grt=grts,
                srcs=srcs,
                mthds=mthds,
                metrics=metrics,
                runby=runby,
                emin=emin,
                verbose=verbose,
                **kwargs
            )
    # Format df
    df = pd.DataFrame(df, columns=['groupby', 'group', 'source', 'method', 'metric', 'score'])
    # Remove repeated columns
    df = df.loc[:, df.nunique(dropna=False) > 1]
    return df


def benchmark(
    adata: AnnData,
    net: pd.DataFrame | dict,
    metrics: str | list = ['auc', 'fscore', 'qrank'],
    groupby: str | None = None,
    runby: str = 'expr',
    sfilt: bool = False,
    thr: float = 0.10,
    emin: int = 5,
    verbose: bool = False,
    kws_decouple: dict = dict(),
    **kwargs
):
    """
    Benchmark enrichment methods or networks on a given set of perturbation experiments.

    Parameters
    ----------

    Returns
    -------
    Dataframe containing metric scores.
    """
    # Validate
    assert isinstance(adata, AnnData), 'adata must be anndata.AnnData'
    _validate_obs(adata.obs)
    groupby = _validate_groupby(obs=adata.obs, groupby=groupby, runby=runby)
    assert isinstance(emin, int) and emin > 0, 'emin must be int and > 0'
    # Init default args
    kws_decouple = kws_decouple.copy()
    kws_decouple.setdefault('tmin', 5)
    kws_decouple.setdefault('args', dict())
    # Clean adata
    for col in list(adata.obsm.keys()):
        if col.startswith('score_') or col.startswith('padj_'):
            del adata.obsm[col]
    # Run benchmark per net
    if isinstance(net, pd.DataFrame):
        m = f'benchmark - running benchmark for one network'
        _log(m, level='info', verbose=verbose)
        # Process
        net = prune(features=adata.var_names, net=net, tmin=kws_decouple['tmin'], verbose=verbose)
        adata, net = _filter(adata=adata, net=net, sfilt=sfilt, verbose=verbose)
        _sign(adata=adata)
        # Run benchmark
        decouple(data=adata, net=net, verbose=verbose, layer='tmp', **kws_decouple)
        df = _eval_scores(
            adata=adata,
            groupby=groupby,
            runby=runby,
            metrics=metrics,
            thr=thr,
            emin=emin,
            verbose=verbose,
            **kwargs
        )
    else:
        m = f'benchmark - running benchmark for multiple networks'
        _log(m, level='info', verbose=verbose)
        df = []
        for n_name in net:
            nnet = net[n_name]
            nnet = prune(features=adata.var_names, net=nnet, tmin=kws_decouple['tmin'], verbose=verbose)
            ndata, nnet = _filter(adata=adata, net=nnet, sfilt=sfilt, verbose=verbose)
            _sign(adata=ndata)
            # Run benchmark
            decouple(data=ndata, net=nnet, verbose=verbose, layer='tmp', **kws_decouple)
            tmp = _eval_scores(
                adata=ndata,
                groupby=groupby,
                runby=runby,
                metrics=metrics,
                thr=thr,
                emin=emin,
                verbose=verbose,
                **kwargs
            )
            tmp.insert(0, 'net', n_name)
            df.append(tmp)
        df = pd.concat(df)
    return df
